from __future__ import annotations

import uuid
from copy import copy, deepcopy
from typing import Dict, Union
from unittest import mock
from unittest.mock import MagicMock, Mock  # noqa: TID251 # FIXME CoP
from uuid import UUID, uuid4

import pytest

import great_expectations as gx
import great_expectations.exceptions as gx_exceptions
import great_expectations.expectations as gxe
from great_expectations import __version__ as ge_version
from great_expectations import get_context
from great_expectations.analytics.events import (
    ExpectationSuiteExpectationCreatedEvent,
    ExpectationSuiteExpectationDeletedEvent,
    ExpectationSuiteExpectationUpdatedEvent,
)
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.serdes import _IdentifierBundle
from great_expectations.data_context import AbstractDataContext
from great_expectations.data_context.data_context.context_factory import set_context
from great_expectations.data_context.store.expectations_store import ExpectationsStore
from great_expectations.exceptions import InvalidExpectationConfigurationError
from great_expectations.expectations import (
    ExpectColumnValuesToBeUnique,
    ExpectColumnValuesToNotBeNull,
)
from great_expectations.expectations.expectation import Expectation
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.render import RenderedAtomicContent, RenderedAtomicValue


@pytest.fixture
def fake_expectation_suite_name() -> str:
    return "test_expectation_suite_name"


@pytest.fixture
def expect_column_values_to_be_in_set_col_a_with_meta() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "a",
            "value_set": [1, 2, 3],
        },
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def expect_column_values_to_be_in_set_col_a_with_meta_dict() -> dict:
    # Note, value_set is distinct to ensure this is treated as a different expectation
    return {
        "type": "expect_column_values_to_be_in_set",
        "kwargs": {
            "column": "a",
            "value_set": [1, 2, 3, 4, 5],
        },
        "meta": {"notes": "This is an expectation."},
    }


@pytest.fixture
def bad_expectation_dict() -> dict:
    return {
        "type": "expect_stuff_not_to_go_well",
        "kwargs": {},
        "meta": {"notes": "this_should_explode"},
    }


@pytest.fixture
def empty_suite_with_meta(fake_expectation_suite_name: str) -> ExpectationSuite:
    return ExpectationSuite(
        name=fake_expectation_suite_name,
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def suite_with_single_expectation(
    expect_column_values_to_be_in_set_col_a_with_meta: ExpectationConfiguration,
    empty_suite_with_meta: ExpectationSuite,
) -> ExpectationSuite:
    empty_suite_with_meta.add_expectation_configuration(
        expect_column_values_to_be_in_set_col_a_with_meta
    )
    return empty_suite_with_meta


class TestInit:
    """Tests related to ExpectationSuite.__init__()"""

    @pytest.mark.unit
    def test_expectation_suite_init_defaults(
        self,
        empty_data_context: AbstractDataContext,
        fake_expectation_suite_name: str,
    ):
        suite = ExpectationSuite(name=fake_expectation_suite_name)

        default_meta = {"great_expectations_version": ge_version}

        assert suite.name == fake_expectation_suite_name
        assert suite.expectations == []
        assert suite.suite_parameters == {}
        assert suite.meta == default_meta
        assert suite.id is None

    @pytest.mark.unit
    def test_expectation_suite_init_overrides(
        self,
        empty_data_context: AbstractDataContext,
        fake_expectation_suite_name: str,
        expect_column_values_to_be_in_set_col_a_with_meta: ExpectationConfiguration,
    ):
        class DummyExecutionEngine:
            pass

        test_suite_parameters = {"$PARAMETER": "test_suite_parameters"}
        default_meta = {"great_expectations_version": ge_version}
        test_meta_base = {"test_key": "test_value"}
        test_meta = {**default_meta, **test_meta_base}
        test_id = "test_id"

        suite = ExpectationSuite(
            name=fake_expectation_suite_name,
            expectations=[expect_column_values_to_be_in_set_col_a_with_meta],
            suite_parameters=test_suite_parameters,
            meta=test_meta,
            id=test_id,
        )
        assert suite.name == fake_expectation_suite_name
        assert suite.expectation_configurations == [
            expect_column_values_to_be_in_set_col_a_with_meta
        ]
        assert suite.suite_parameters == test_suite_parameters
        assert suite.meta == test_meta
        assert suite.id == test_id

    @pytest.mark.unit
    def test_expectation_suite_init_overrides_expectations_dict_and_obj(
        self,
        empty_data_context: AbstractDataContext,
        fake_expectation_suite_name: str,
        expect_column_values_to_be_in_set_col_a_with_meta_dict: dict,
        expect_column_values_to_be_in_set_col_a_with_meta: ExpectationConfiguration,
    ):
        """What does this test and why?

        The expectations param of ExpectationSuite takes a list of ExpectationConfiguration or dicts and both can be provided at the same time. We need to make sure they both show up as expectation configurations in the instantiated ExpectationSuite.
        """  # noqa: E501 # FIXME CoP

        test_expectations_input = [
            expect_column_values_to_be_in_set_col_a_with_meta_dict,
            expect_column_values_to_be_in_set_col_a_with_meta,
        ]

        suite = ExpectationSuite(
            name=fake_expectation_suite_name,
            expectations=test_expectations_input,  # type: ignore[arg-type] # FIXME CoP
        )
        assert suite.name == fake_expectation_suite_name

        test_expected_expectations = [
            ExpectationConfiguration(**expect_column_values_to_be_in_set_col_a_with_meta_dict),
            expect_column_values_to_be_in_set_col_a_with_meta,
        ]
        assert len(suite.expectations) == 2
        assert suite.expectation_configurations == test_expected_expectations

    @pytest.mark.unit
    def test_bad_expectation_configs_are_skipped(
        self,
        bad_expectation_dict: dict,
        expect_column_values_to_be_in_set_col_a_with_meta_dict: dict,
    ):
        suite = ExpectationSuite(
            name="test_suite",
            expectations=[
                bad_expectation_dict,
                expect_column_values_to_be_in_set_col_a_with_meta_dict,
            ],
        )
        assert len(suite.expectations) == 1
        assert (
            suite.expectations[0].expectation_type
            == expect_column_values_to_be_in_set_col_a_with_meta_dict["type"]
        )

    @pytest.mark.unit
    def test_expectation_suite_init_overrides_non_json_serializable_meta(
        self,
        fake_expectation_suite_name: str,
    ):
        """What does this test and why?

        meta field overrides need to be json serializable, if not we raise an exception.
        """

        class NotSerializable:
            def __dict__(self):
                raise NotImplementedError

        test_meta = {"this_is_not_json_serializable": NotSerializable()}

        with pytest.raises(InvalidExpectationConfigurationError) as e:
            ExpectationSuite(
                name=fake_expectation_suite_name,
                meta=test_meta,  # type: ignore[arg-type] # FIXME CoP
            )
        assert "is of type NotSerializable which cannot be serialized to json" in str(e.value)


class TestCRUDMethods:
    """Tests related to the 1.0 CRUD API."""

    expectation_suite_name = "test-suite"

    @pytest.fixture
    def expectation(self) -> gxe.ExpectColumnValuesToBeInSet:
        return gxe.ExpectColumnValuesToBeInSet(
            column="a",
            value_set=[1, 2, 3],
            result_format="BASIC",
        )

    @pytest.mark.unit
    def test_instantiate_suite_with_expectations(self, expectation):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)
        columns = ["a", "b", "c", "d", "e"]
        expectations = [expectation.copy(update={"column": column}) for column in columns]
        suite = ExpectationSuite(name=self.expectation_suite_name, expectations=expectations)
        assert suite.expectations == expectations

    @pytest.mark.unit
    def test_instantiate_suite_fails_with_expectations_with_ids(self, expectation):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)
        columns = ["a", "b", "c", "d", "e"]
        expectations = [
            expectation.copy(update={"column": column, "id": uuid4()}) for column in columns
        ]
        with pytest.raises(
            ValueError,
            match="Expectations in parameter `expectations` must not belong to another ExpectationSuite.",  # noqa: E501 # FIXME CoP
        ):
            ExpectationSuite(name=self.expectation_suite_name, expectations=expectations)

    @pytest.mark.unit
    def test_add_success_with_saved_suite(self, expectation):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)
        suite = ExpectationSuite(name=self.expectation_suite_name)

        with mock.patch.object(ExpectationSuite, "_submit_expectation_created_event"):
            created_expectation = suite.add_expectation(expectation=expectation)

        assert created_expectation == context.expectations_store.add_expectation.return_value
        context.expectations_store.add_expectation.assert_called_once_with(
            suite=suite, expectation=expectation
        )

    @pytest.mark.unit
    def test_add_success_with_unsaved_suite(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = False
        set_context(project=context)
        suite = ExpectationSuite(name=self.expectation_suite_name)

        created_expectation = suite.add_expectation(expectation=expectation)

        assert created_expectation == expectation
        assert len(suite.expectations) == 1

        # expect that adding an expectation to this suite doesnt have the side effect of
        # persisting the suite to the data context
        context.expectations_store.set.assert_not_called()

    @pytest.mark.unit
    def test_add_doesnt_duplicate_when_expectation_already_exists(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = True
        set_context(project=context)
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation],
        )

        suite.add_expectation(expectation=expectation)

        assert len(suite.expectations) == 1
        context.expectations_store.update.assert_not_called()

    @pytest.mark.unit
    def test_add_doesnt_duplicate_when_expectation_differs_rendered_content(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = True
        set_context(project=context)
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation],
        )

        dup_expectation = deepcopy(expectation)
        dup_expectation.render()

        assert expectation.rendered_content != dup_expectation.rendered_content

        suite.add_expectation(expectation=dup_expectation)

        assert len(suite.expectations) == 1
        context.expectations_store.update.assert_not_called()

    @pytest.mark.unit
    def test_add_doesnt_duplicate_when_expectation_differs_meta(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = True
        set_context(project=context)
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation],
        )

        dup_expectation = deepcopy(expectation)
        dup_expectation.meta = {"author": "Guido van Rossum"}

        assert expectation.meta != dup_expectation.meta

        suite.add_expectation(expectation=dup_expectation)

        assert len(suite.expectations) == 1
        context.expectations_store.update.assert_not_called()

    @pytest.mark.unit
    def test_add_doesnt_duplicate_when_expectation_differs_notes(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = True
        set_context(project=context)
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation],
        )

        dup_expectation = deepcopy(expectation)
        dup_expectation.notes = "These are my notes!"

        assert expectation.notes != dup_expectation.notes

        suite.add_expectation(expectation=dup_expectation)

        assert len(suite.expectations) == 1
        context.expectations_store.update.assert_not_called()

    @pytest.mark.unit
    def test_add_doesnt_duplicate_when_expectation_differs_id(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = True
        set_context(project=context)
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation],
        )
        suite.expectations[0].id = str(uuid.uuid4())

        dup_expectation = deepcopy(expectation)
        dup_expectation.id = None

        assert expectation.id != dup_expectation.id

        suite.add_expectation(expectation=dup_expectation)

        assert len(suite.expectations) == 1
        context.expectations_store.update.assert_not_called()

    @pytest.mark.unit
    def test_add_doesnt_mutate_suite_when_save_fails(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.add_expectation.side_effect = (
            ConnectionError()
        )  # arbitrary exception
        context.expectations_store.has_key.return_value = True
        set_context(project=context)
        suite = ExpectationSuite(
            name="test-suite",
        )

        with pytest.raises(ConnectionError):  # exception type isn't important
            suite.add_expectation(expectation=expectation)

        assert len(suite.expectations) == 0, "Expectation must not be added to Suite."

    @pytest.mark.unit
    def test_add_success_when_attributes_are_identical(self):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)

        parameters = {"column": "passenger_count"}
        expectation_a = ExpectColumnValuesToBeUnique(**parameters)
        expectation_b = ExpectColumnValuesToNotBeNull(**parameters)

        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation_a],
        )

        with mock.patch.object(ExpectationSuite, "_submit_expectation_created_event"):
            suite.add_expectation(expectation=expectation_b)

        assert len(suite.expectations) == 2

    @pytest.mark.unit
    def test_delete_success_with_saved_suite(self, expectation):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)
        context.expectations_store.has_key.return_value = True
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation],
        )

        deleted_expectation = suite.delete_expectation(expectation=expectation)

        assert deleted_expectation == expectation
        assert suite.expectations == []

        # expect that the data context is kept in sync with the mutation
        context.expectations_store.delete_expectation.assert_called_once_with(
            suite=suite, expectation=expectation
        )

    @pytest.mark.unit
    def test_delete_success_with_unsaved_suite(self, expectation):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)
        context.expectations_store.has_key.return_value = False
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
            expectations=[expectation],
        )

        deleted_expectation = suite.delete_expectation(expectation=expectation)

        assert deleted_expectation == expectation
        assert suite.expectations == []
        # expect that deleting an expectation from this suite doesnt have the side effect of
        # persisting the suite to the data context
        context.expectations_store.delete_expectation.assert_not_called()

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "input_kwargs",
        [
            pytest.param({"id": str(uuid.uuid4())}, id="id"),
            pytest.param({"meta": {"author": "Alexandre Dumas"}}, id="meta"),
            pytest.param({"notes": "Just some thoughts!"}, id="notes"),
        ],
    )
    def test_delete_success_with_equalish_expectation(self, input_kwargs: dict):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)

        expectation = gxe.ExpectColumnValuesToBeInSet(column="a", value_set=[1, 2, 3])
        suite = ExpectationSuite(
            name="test-suite",
            expectations=[expectation],
        )

        suite.delete_expectation(
            expectation=gxe.ExpectColumnValuesToBeInSet(
                column="a", value_set=[1, 2, 3], **input_kwargs
            )
        )
        assert suite.expectations == []

    @pytest.mark.unit
    def test_delete_fails_when_expectation_is_not_found(self, expectation):
        context = Mock(spec=AbstractDataContext)
        set_context(project=context)
        suite = ExpectationSuite(
            name="test-suite",
        )

        with pytest.raises(KeyError, match="No matching expectation was found."):
            suite.delete_expectation(expectation=expectation)

        context.expectations_store.delete_expectation.assert_not_called()

    @pytest.mark.unit
    def test_delete_doesnt_mutate_suite_when_save_fails(self, expectation):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = True
        context.expectations_store.delete_expectation.side_effect = (
            ConnectionError()
        )  # arbitrary exception
        set_context(project=context)
        suite = ExpectationSuite(
            name="test-suite",
            expectations=[expectation],
        )

        with pytest.raises(ConnectionError):  # exception type isn't important
            suite.delete_expectation(expectation=expectation)

        assert len(suite.expectations) == 1, "Expectation must still be in Suite."

    @pytest.mark.unit
    def test_save_success(self):
        context = Mock(spec=AbstractDataContext)
        context.expectations_store.has_key.return_value = True
        set_context(project=context)
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
        )
        store_key = context.expectations_store.get_key.return_value

        suite.save()

        # expect that the data context is kept in sync
        context.expectations_store.update.assert_called_once_with(key=store_key, value=suite)

    @pytest.mark.unit
    def test_save_before_add_raises(self):
        gx.get_context(mode="ephemeral")
        suite = ExpectationSuite(
            name=self.expectation_suite_name,
        )

        with pytest.raises(gx_exceptions.ExpectationSuiteNotAddedError):
            suite.save()

    @pytest.mark.filesystem
    def test_filesystem_context_update_suite_adds_ids(self, empty_data_context, expectation):
        context = empty_data_context
        self._test_update_suite_adds_ids(context, expectation)

    @pytest.mark.cloud
    def test_cloud_context_update_suite_adds_ids(self, empty_cloud_context_fluent, expectation):
        context = empty_cloud_context_fluent
        self._test_update_suite_adds_ids(context, expectation)

    def _test_update_suite_adds_ids(self, context, expectation):
        suite_name = "test-suite"
        suite = ExpectationSuite(suite_name)
        suite = context.suites.add(suite)
        uuid_to_test = suite.id
        try:
            UUID(uuid_to_test)
        except TypeError:
            pytest.fail(f"Expected UUID in ExpectationSuite.id, found {uuid_to_test}")
        expectation.id = None
        suite.add_expectation(expectation)
        expectation = copy(expectation)
        expectation.column = "foo"
        suite.add_expectation(expectation)
        assert len(suite.expectations) == 2
        for expectation in suite.expectations:
            uuid_to_test = expectation.id
            try:
                UUID(uuid_to_test)
            except TypeError:
                pytest.fail(f"Expected UUID in ExpectationConfiguration.id, found {uuid_to_test}")

    @pytest.mark.unit
    def test_suite_add_expectation_doesnt_allow_adding_an_expectation_with_id(self, expectation):
        suite = ExpectationSuite("test-suite")
        provided_id = "6b3f003d-d97b-4649-ba23-3f4e30986297"
        expectation.id = provided_id

        with pytest.raises(
            RuntimeError,
            match="Cannot add Expectation because it already belongs to an ExpectationSuite.",
        ):
            suite.add_expectation(expectation)

    @pytest.mark.cloud
    def test_cloud_expectation_can_be_saved_after_added(
        self, empty_cloud_context_fluent, expectation
    ):
        context = empty_cloud_context_fluent
        self._test_expectation_can_be_saved_after_added(context, expectation)

    @pytest.mark.filesystem
    def test_filesystem_expectation_can_be_saved_after_added(self, empty_data_context, expectation):
        context = empty_data_context
        self._test_expectation_can_be_saved_after_added(context, expectation)

    def _test_expectation_can_be_saved_after_added(self, context, expectation):
        suite_name = "test-suite"
        suite = ExpectationSuite(suite_name)
        suite = context.suites.add(suite)
        suite.add_expectation(expectation)
        updated_column_name = "foo"
        assert expectation.column != updated_column_name
        expectation.column = updated_column_name
        expectation.save()
        assert suite.expectations[0].column == updated_column_name
        suite = context.suites.get(suite_name)
        assert len(suite.expectations) == 1
        assert suite.expectations[0].column == updated_column_name

    @pytest.mark.cloud
    def test_cloud_expectation_can_be_saved_after_update(
        self, empty_cloud_context_fluent, expectation
    ):
        context = empty_cloud_context_fluent
        self._test_expectation_can_be_saved_after_update(context, expectation)

    @pytest.mark.filesystem
    def test_filesystem_expectation_can_be_saved_after_update(
        self, empty_data_context, expectation
    ):
        context = empty_data_context
        self._test_expectation_can_be_saved_after_update(context, expectation)

    def _test_expectation_can_be_saved_after_update(self, context, expectation):
        suite_name = "test-suite"
        suite = ExpectationSuite(suite_name, expectations=[expectation])
        suite = context.suites.add(suite)
        expectation = suite.expectations[0]
        updated_column_name = "foo"
        expectation.column = updated_column_name
        expectation.save()
        assert suite.expectations[0].column == updated_column_name
        suite = context.suites.get(suite_name)
        assert len(suite.expectations) == 1
        assert suite.expectations[0].column == updated_column_name

    @pytest.mark.filesystem
    def test_expectation_save_only_persists_single_change(self, empty_data_context):
        context = empty_data_context
        suite_name = "test-suite"
        expectations = [
            gxe.ExpectColumnValuesToBeInSet(
                column="a",
                value_set=[1, 2, 3],
            ),
            gxe.ExpectColumnValuesToBeInSet(
                column="b",
                value_set=[4, 5, 6],
            ),
        ]

        suite = ExpectationSuite(
            name=suite_name,
            expectations=expectations,
        )
        context.suites.add(suite)

        assert suite.expectations[0].column == "a"
        assert suite.expectations[1].column == "b"

        # Change the column names of both expectations
        expectation = suite.expectations[0]
        other_expectation = suite.expectations[1]
        expectation.column = "c"
        other_expectation.column = "d"

        # Only persist change made to first expectation
        expectation.save()

        updated_suite = context.suites.get(name=suite_name)

        assert updated_suite.expectations[0].column == "c"
        assert updated_suite.expectations[1].column == "b"

    @pytest.mark.cloud
    def test_expectation_save_callback_can_come_from_any_copy_of_a_suite(
        self, empty_cloud_context_fluent
    ):
        """Equivalent calls to ExpectationSuite._save_expectation from different copies of a
        single ExpectationSuite must produce equivalent side effects.

        In some cases, the ExpectationsStore replaces Expectations from a given suite with Expectations
        from another copy of the same suite, in order to keep the ExpectationSuite in memory up to date
        with the remote ExpectationSuite. ExpectationSuite._save_expectation (and the corresponding logic
        the suite uses within the ExpectationsStore) must work equivalently regardless of which Suite instance
        it belongs to.
        """  # noqa: E501 # FIXME CoP
        # Arrange
        context = empty_cloud_context_fluent
        suite_name = "test-suite"
        suite_a = ExpectationSuite(name=suite_name)
        column_name = "a"
        updated_column_name = "foo"
        expectations = [
            gxe.ExpectColumnValuesToBeInSet(
                column=column_name,
                value_set=[1, 2, 3],
            ),
            gxe.ExpectColumnValuesToBeInSet(
                column="b",
                value_set=[4, 5, 6],
            ),
        ]
        for expectation in expectations:
            suite_a.add_expectation(expectation)

        context.suites.add(suite_a)

        suite_b = context.suites.get(name=suite_name)

        # Act
        suite_a.expectations = suite_b.expectations
        expectation = suite_a.expectations[0]
        # the following assert is the method equivalent of `is`
        assert expectation._save_callback == suite_b._save_expectation
        assert expectation.column == column_name
        expectation.column = updated_column_name
        expectation.save()

        # Assert
        fetched_suite = context.suites.get(name=suite_name)
        assert (
            suite_a.expectations[0].column
            == fetched_suite.expectations[0].column
            == updated_column_name
        )

    @pytest.mark.unit
    def test_expectation_suite_name_can_be_updated(self, empty_cloud_context_fluent):
        """Expect that ExpectationSuite.name can be updated directly"""
        suite = ExpectationSuite(name=self.expectation_suite_name)
        new_name = "updated name"
        suite.name = "updated name"
        assert suite.name == new_name


class TestSuiteParameterOptions:
    """Tests around the suite_parameter_options property of ExpectationSuites.

    Note: suite_parameter_options is currently a sorted tuple, but doesn't necessarily have to be
    """

    SUITE_PARAMETER_VALUE_SET = "my_value_set"
    SUITE_PARAMETER_MIN = "my_min"
    SUITE_PARAMETER_MAX = "my_max"

    @pytest.fixture
    def expectation_suite(self) -> ExpectationSuite:
        get_context(mode="ephemeral")
        return ExpectationSuite("test-suite")

    @pytest.fixture
    def expectation_with_suite_parameter(
        self,
    ) -> Expectation:
        return gxe.ExpectColumnDistinctValuesToBeInSet(
            column="a", value_set={"$PARAMETER": self.SUITE_PARAMETER_VALUE_SET}
        )

    @pytest.fixture
    def expectation_with_duplicate_suite_parameter(
        self,
    ) -> Expectation:
        return gxe.ExpectColumnDistinctValuesToContainSet(
            column="a", value_set={"$PARAMETER": self.SUITE_PARAMETER_VALUE_SET}
        )

    @pytest.fixture
    def expectation_with_multiple_suite_parameters(
        self,
    ) -> Expectation:
        return gxe.ExpectColumnValuesToBeBetween(
            column="c",
            min_value={"$PARAMETER": self.SUITE_PARAMETER_MIN},
            max_value={"$PARAMETER": self.SUITE_PARAMETER_MAX},
        )

    @pytest.fixture
    def expectation_without_suite_parameters(
        self,
    ) -> Expectation:
        return gxe.ExpectColumnDistinctValuesToBeInSet(column="d", value_set=[1, 2])

    @pytest.mark.unit
    def test_empty_suite(self, expectation_suite: ExpectationSuite):
        assert expectation_suite.suite_parameter_options == tuple()

    @pytest.mark.unit
    def test_expectations_but_no_evaluation_params(
        self,
        expectation_suite: ExpectationSuite,
        expectation_without_suite_parameters: Expectation,
    ):
        expectation_suite.add_expectation(expectation_without_suite_parameters)

        assert expectation_suite.suite_parameter_options == tuple()

    @pytest.mark.unit
    def test_expectation_with_suite_parameter(
        self,
        expectation_suite: ExpectationSuite,
        expectation_with_suite_parameter: Expectation,
    ):
        expectation_suite.add_expectation(expectation_with_suite_parameter)

        assert expectation_suite.suite_parameter_options == (self.SUITE_PARAMETER_VALUE_SET,)

    @pytest.mark.unit
    def test_duplicate_suite_parameters_only_show_once(
        self,
        expectation_suite: ExpectationSuite,
        expectation_with_suite_parameter: Expectation,
        expectation_with_duplicate_suite_parameter: Expectation,
    ):
        expectation_suite.add_expectation(expectation_with_suite_parameter)
        expectation_suite.add_expectation(expectation_with_duplicate_suite_parameter)

        assert expectation_suite.suite_parameter_options == (self.SUITE_PARAMETER_VALUE_SET,)

    @pytest.mark.unit
    def test_multiple_suite_parameters_on_one_expectation(
        self,
        expectation_suite: ExpectationSuite,
        expectation_with_multiple_suite_parameters: Expectation,
    ):
        expectation_suite.add_expectation(expectation_with_multiple_suite_parameters)

        assert expectation_suite.suite_parameter_options == (
            self.SUITE_PARAMETER_MAX,
            self.SUITE_PARAMETER_MIN,
        )

    @pytest.mark.unit
    def test_multiple_suite_parameters_across_multiple_expectation(
        self,
        expectation_suite: ExpectationSuite,
        expectation_with_suite_parameter: Expectation,
        expectation_with_multiple_suite_parameters: Expectation,
    ):
        expectation_suite.add_expectation(expectation_with_suite_parameter)
        expectation_suite.add_expectation(expectation_with_multiple_suite_parameters)

        assert expectation_suite.suite_parameter_options == (
            self.SUITE_PARAMETER_MAX,
            self.SUITE_PARAMETER_MIN,
            self.SUITE_PARAMETER_VALUE_SET,
        )


class TestEqDunder:
    @pytest.mark.unit
    def test_equality_to_unsupported_class_is_false(
        self, suite_with_single_expectation: ExpectationSuite
    ):
        """If we are not comparing to an ExpectationSuite or dict then we should return False."""

        class UnsupportedClass:
            pass

        return_value = suite_with_single_expectation == UnsupportedClass()
        assert not return_value

    @pytest.mark.unit
    def test_expectation_suite_equality_single_expectation_true(
        self, suite_with_single_expectation: ExpectationSuite
    ):
        different_but_equivalent_suite = deepcopy(suite_with_single_expectation)
        assert suite_with_single_expectation == different_but_equivalent_suite
        assert different_but_equivalent_suite == suite_with_single_expectation

        assert not suite_with_single_expectation != different_but_equivalent_suite
        assert not different_but_equivalent_suite != suite_with_single_expectation

    @pytest.mark.parametrize(
        "attribute,new_value",
        [
            pytest.param("name", "different_name"),
            pytest.param(
                "data_context",
                MagicMock(),
                marks=pytest.mark.xfail(
                    strict=True,
                    raises=AssertionError,
                    reason="Currently data_context is not considered in ExpectationSuite equality",
                ),
            ),
            pytest.param("expectations", []),
            pytest.param("suite_parameters", {"different": "suite_parameters"}),
            pytest.param("meta", {"notes": "Different meta."}),
            pytest.param(
                "id",
                "different_id",
                marks=pytest.mark.xfail(
                    strict=True,
                    raises=AssertionError,
                    reason="Currently id is not considered in ExpectationSuite equality",
                ),
            ),
        ],
    )
    @pytest.mark.unit
    def test_expectation_suite_equality_false(
        self,
        attribute: str,
        new_value: Union[str, Dict[str, str]],
        suite_with_single_expectation: ExpectationSuite,
    ):
        different_but_equivalent_suite = deepcopy(suite_with_single_expectation)

        setattr(different_but_equivalent_suite, attribute, new_value)

        assert suite_with_single_expectation != different_but_equivalent_suite
        assert different_but_equivalent_suite != suite_with_single_expectation
        assert suite_with_single_expectation != different_but_equivalent_suite
        assert different_but_equivalent_suite != suite_with_single_expectation


class TestExpectationsAreEqualish:
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "expectation_a,expectation_b",
        [
            pytest.param(
                gxe.ExpectColumnValuesToBeInSet(
                    column="a",
                    value_set=[1, 2, 3],
                    result_format="BASIC",
                ),
                gxe.ExpectColumnValuesToBeInSet(
                    column="a",
                    value_set=[1, 2, 3],
                    result_format="BASIC",
                ),
                id="same args passed in",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotBeNull(column="a", id=str(uuid4())),
                gxe.ExpectColumnValuesToNotBeNull(column="a", id=str(uuid4())),
                id="different ids",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotBeNull(column="a", rendered_content=[]),
                gxe.ExpectColumnValuesToNotBeNull(
                    column="a",
                    rendered_content=[
                        RenderedAtomicContent(
                            name="atomic.prescriptive.summary",
                            value=RenderedAtomicValue(template="Render this string!"),
                            value_type="StringValueType",
                        )
                    ],
                ),
                id="different rendered_content",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotBeNull(column="a", notes="Note ABC"),
                gxe.ExpectColumnValuesToNotBeNull(column="a", notes="Note 123"),
                id="different notes",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotBeNull(column="a", meta={"warehouse": "theirs"}),
                gxe.ExpectColumnValuesToNotBeNull(column="a", meta={"warehouse": "ours"}),
                id="different meta",
            ),
        ],
    )
    def test_equalish(self, expectation_a, expectation_b):
        assert ExpectationSuite._expectations_are_equalish(expectation_a, expectation_b)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "expectation_a,expectation_b",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotBeNull(column="a"),
                gxe.ExpectColumnValuesToBeUnique(column="b"),
                id="different expectation, different args passed in",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotBeNull(column="a"),
                gxe.ExpectColumnValuesToBeUnique(column="a"),
                id="different expectation, same args passed in",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotBeNull(column="a"),
                gxe.ExpectColumnValuesToNotBeNull(column="b"),
                id="same expectation, different args passed in",
            ),
        ],
    )
    def test_not_equalish(self, expectation_a, expectation_b):
        assert not ExpectationSuite._expectations_are_equalish(expectation_a, expectation_b)


# ### Below this line are mainly existing tests and fixtures that we are in the process of cleaning up  # noqa: E501 # FIXME CoP


@pytest.fixture
def exp2() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3]},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp3() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "b",
            "value_set": [-1, -2, -3],
        },
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp4() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "b",
            "value_set": [1, 2, 3],
        },
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def column_pair_expectation() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_pair_values_to_be_in_set",
        kwargs={
            "column_A": "1",
            "column_B": "b",
            "value_pairs_set": [(1, 1), (2, 2)],
        },
    )


@pytest.fixture
def table_exp1() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_table_columns_to_match_ordered_list",
        kwargs={"column_list": ["a", "b", "c"]},
    )


@pytest.fixture
def table_exp2() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_table_row_count_to_be_between",
        kwargs={"min_value": 0, "max_value": 1},
    )


@pytest.fixture
def table_exp3() -> ExpectationConfiguration:
    return ExpectationConfiguration(type="expect_table_row_count_to_equal", kwargs={"value": 1})


@pytest.fixture
def empty_suite() -> ExpectationSuite:
    return ExpectationSuite(
        name="warning",
        expectations=[],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def suite_with_table_and_column_expectations(
    expect_column_values_to_be_in_set_col_a_with_meta,
    exp2,
    exp3,
    exp4,
    column_pair_expectation,
    table_exp1,
    table_exp2,
    table_exp3,
) -> ExpectationSuite:
    suite = ExpectationSuite(
        name="warning",
        expectations=[
            expect_column_values_to_be_in_set_col_a_with_meta,
            exp2,
            exp3,
            exp4,
            column_pair_expectation,
            table_exp1,
            table_exp2,
            table_exp3,
        ],
        meta={"notes": "This is an expectation suite."},
    )
    return suite


@pytest.fixture
def baseline_suite(expect_column_values_to_be_in_set_col_a_with_meta, exp2) -> ExpectationSuite:
    return ExpectationSuite(
        name="warning",
        expectations=[expect_column_values_to_be_in_set_col_a_with_meta, exp2],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.mark.unit
def test_expectation_suite_copy(baseline_suite):
    suite_copy = copy(baseline_suite)
    assert suite_copy == baseline_suite
    suite_copy.name = "blarg!"
    assert baseline_suite.name != "blarg"  # copy on primitive properties shouldn't propagate


@pytest.mark.unit
def test_expectation_suite_deepcopy(baseline_suite):
    suite_deepcopy = deepcopy(baseline_suite)
    assert suite_deepcopy == baseline_suite
    suite_deepcopy.name = "blarg!"
    assert baseline_suite.name != "blarg"  # copy on primitive properties shouldn't propagate
    suite_deepcopy.expectation_configurations[0].meta["notes"] = "a different note"
    # deepcopy on deep attributes does not propagate
    assert baseline_suite.expectation_configurations[0].meta["notes"] == "This is an expectation."


@pytest.mark.unit
def test_suite_without_metadata_includes_ge_version_metadata_if_none_is_provided():
    suite = ExpectationSuite("foo")
    assert "great_expectations_version" in suite.meta


@pytest.mark.unit
def test_suite_does_not_overwrite_existing_version_metadata():
    suite = ExpectationSuite(
        "foo",
        meta={"great_expectations_version": "0.0.0"},
    )
    assert "great_expectations_version" in suite.meta
    assert suite.meta["great_expectations_version"] == "0.0.0"


@pytest.mark.unit
def test_suite_with_metadata_includes_ge_version_metadata(baseline_suite):
    assert "great_expectations_version" in baseline_suite.meta


class DataContextSendUsageMessageSpy:
    def __init__(self):
        self.messages = []

    def send_usage_message(
        self,
        event,
        event_payload,
        success,
    ):
        self.messages.append(
            {
                "event": event,
                "event_payload": event_payload,
                "success": success,
            }
        )


@pytest.mark.unit
def test_add_expectation_fails_validation(empty_suite_with_meta: ExpectationSuite):
    suite = empty_suite_with_meta

    expectation_type = "my_fake_expectation"
    kwargs = {"foo": "bar"}
    expectation_configuration = ExpectationConfiguration(
        type=expectation_type,
        kwargs=kwargs,
    )

    with pytest.raises(gx_exceptions.InvalidExpectationConfigurationError) as e:
        suite.add_expectation_configuration(expectation_configuration)

    assert f"{expectation_type} not found" in str(e)


class TestExpectationSuiteAnalytics:
    @pytest.fixture
    def empty_suite(self, in_memory_runtime_context) -> ExpectationSuite:
        return in_memory_runtime_context.suites.add(ExpectationSuite(name="my_suite"))

    @pytest.fixture
    def expect_column_values_to_be_between(self) -> Expectation:
        return gxe.ExpectColumnValuesToBeBetween(column="passenger_count", min_value=1, max_value=6)

    @pytest.mark.unit
    def test_add_expectation_emits_event(self, empty_suite, expect_column_values_to_be_between):
        suite = empty_suite
        expectation = expect_column_values_to_be_between

        with mock.patch("great_expectations.core.expectation_suite.submit_event") as mock_submit:
            _ = suite.add_expectation(expectation)

        mock_submit.assert_called_once_with(
            event=ExpectationSuiteExpectationCreatedEvent(
                expectation_id=mock.ANY,
                expectation_suite_id=mock.ANY,
                expectation_type="expect_column_values_to_be_between",
                custom_exp_type=False,
            )
        )

    @pytest.mark.unit
    def test_add_custom_expectation_emits_event(self, empty_suite):
        suite = empty_suite

        class ExpectColumnValuesToBeBetweenOneAndTen(gxe.ExpectColumnValuesToBeBetween):
            min_value: int = 1
            max_value: int = 10

        expectation = ExpectColumnValuesToBeBetweenOneAndTen(column="passenger_count")

        with mock.patch("great_expectations.core.expectation_suite.submit_event") as mock_submit:
            _ = suite.add_expectation(expectation)

        mock_submit.assert_called_once_with(
            event=ExpectationSuiteExpectationCreatedEvent(
                expectation_id=mock.ANY,
                expectation_suite_id=mock.ANY,
                expectation_type=mock.ANY,
                custom_exp_type=True,
            )
        )

        # Due to anonymizer randomization, we can't assert the exact expectation type
        # We can however assert that it has been hashed
        expectation_type = mock_submit.call_args.kwargs["event"].expectation_type
        assert not expectation_type.startswith("expect_")

    @pytest.mark.unit
    def test_delete_expectation_emits_event(self, empty_suite, expect_column_values_to_be_between):
        suite = empty_suite
        expectation = expect_column_values_to_be_between

        suite.add_expectation(expectation)

        with mock.patch("great_expectations.core.expectation_suite.submit_event") as mock_submit:
            suite.delete_expectation(expectation)

        mock_submit.assert_called_once_with(
            event=ExpectationSuiteExpectationDeletedEvent(
                expectation_id=mock.ANY,
                expectation_suite_id=mock.ANY,
            )
        )

    @pytest.mark.unit
    def test_expectation_save_callback_emits_event(
        self, empty_suite, expect_column_values_to_be_between
    ):
        suite = empty_suite
        expectation = expect_column_values_to_be_between

        expectation = suite.add_expectation(expectation)
        expectation.column = "fare_amount"

        with mock.patch("great_expectations.core.expectation_suite.submit_event") as mock_submit:
            expectation.save()

        mock_submit.assert_called_once_with(
            event=ExpectationSuiteExpectationUpdatedEvent(
                expectation_id=mock.ANY,
                expectation_suite_id=mock.ANY,
            )
        )


@pytest.mark.unit
def test_identifier_bundle_with_existing_id(in_memory_runtime_context):
    context = in_memory_runtime_context
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    suite_id = suite.id

    assert suite.identifier_bundle() == _IdentifierBundle(
        name="my_suite",
        id=suite_id,
    )


@pytest.mark.unit
def test_identifier_bundle_no_id_raises_error():
    _ = gx.get_context(mode="ephemeral")
    suite = ExpectationSuite(name="my_suite", id=None)

    with pytest.raises(gx_exceptions.ResourceFreshnessAggregateError) as e:
        suite.identifier_bundle()

    assert len(e.value.errors) == 1
    assert isinstance(e.value.errors[0], gx_exceptions.ExpectationSuiteNotAddedError)


@pytest.mark.parametrize(
    "id,is_fresh,num_errors",
    [
        pytest.param(str(uuid.uuid4()), True, 0, id="added"),
        pytest.param(None, False, 1, id="not_added"),
    ],
)
@pytest.mark.unit
def test_is_fresh_is_added(
    in_memory_runtime_context, id: str | None, is_fresh: bool, num_errors: int
):
    context = in_memory_runtime_context
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    suite.id = id  # Stores will add an ID but manually overriding for test
    diagnostics = suite.is_fresh()

    assert diagnostics.success is is_fresh
    assert len(diagnostics.errors) == num_errors
    assert all(
        isinstance(err, gx_exceptions.ExpectationSuiteNotAddedError) for err in diagnostics.errors
    )


@pytest.mark.cloud
def test_save_on_suite_updates_rendered_content(
    empty_cloud_context_fluent,
):
    context = empty_cloud_context_fluent

    expectation_1 = gxe.ExpectColumnDistinctValuesToBeInSet(column="a", value_set=[1, 2, 3])
    expectation_2 = gxe.ExpectColumnValuesToBeBetween(column="a", min_value=1, max_value=5)
    suite = context.suites.add(
        suite=ExpectationSuite(
            name="my_suite",
            expectations=[
                expectation_1,
                expectation_2,
            ],
        )
    )

    old_expectation_1_rendered_content = suite.expectations[0].rendered_content
    old_expectation_2_rendered_content = suite.expectations[1].rendered_content

    suite.expectations[0].value_set = [1, 2, 3, 4, 5]
    suite.expectations[1].min_value = 2

    suite.save()

    new_expectation_1_rendered_content = suite.expectations[0].rendered_content
    new_expectation_2_rendered_content = suite.expectations[1].rendered_content

    assert new_expectation_1_rendered_content != old_expectation_1_rendered_content
    assert new_expectation_1_rendered_content[0].value["params"]["value_set"]["value"] == [
        1,
        2,
        3,
        4,
        5,
    ]
    assert new_expectation_2_rendered_content != old_expectation_2_rendered_content
    assert new_expectation_2_rendered_content[0].value["params"]["min_value"]["value"] == 2


@pytest.mark.cloud
def test_save_on_individual_expectation_updates_rendered_content(
    empty_cloud_context_fluent,
):
    context = empty_cloud_context_fluent

    suite = context.suites.add(
        suite=ExpectationSuite(
            name="my_suite",
            expectations=[gxe.ExpectColumnDistinctValuesToBeInSet(column="a", value_set=[1, 2, 3])],
        )
    )

    old_expectation_rendered_content = suite.expectations[0].rendered_content
    suite.expectations[0].value_set = [1, 2, 3, 4, 5]

    suite.expectations[0].save()
    new_expectation_rendered_content = suite.expectations[0].rendered_content

    assert new_expectation_rendered_content != old_expectation_rendered_content
    assert new_expectation_rendered_content[0].value["params"]["value_set"]["value"] == [
        1,
        2,
        3,
        4,
        5,
    ]


@pytest.mark.unit
def test_is_fresh_freshness(in_memory_runtime_context):
    context = in_memory_runtime_context

    suite = context.suites.add(ExpectationSuite(name="my_suite"))

    suite.expectations = [gxe.ExpectColumnDistinctValuesToBeInSet(column="a", value_set=[1, 2, 3])]

    diagnostics = suite.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], gx_exceptions.ExpectationSuiteNotFreshError)


@pytest.mark.unit
def test_is_fresh_fails_on_suite_retrieval(in_memory_runtime_context):
    context = in_memory_runtime_context

    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    context.suites.delete(suite.name)

    diagnostics = suite.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], gx_exceptions.ExpectationSuiteNotFoundError)


@pytest.mark.unit
def test_is_fresh_fails_on_suite_deserialization(in_memory_runtime_context):
    context = in_memory_runtime_context

    suite = context.suites.add(ExpectationSuite(name="my_suite"))

    # Value has changed since initial creation and subsequent retrieval
    suite_dict = {
        "name": "my_suite",
        "expectations": [{"type": "expect_column_values_to_be_between", "kwargs": {}}],
    }
    with mock.patch.object(ExpectationsStore, "get", return_value=suite_dict):
        diagnostics = suite.is_fresh()

    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], gx_exceptions.ExpectationSuiteError)


@pytest.mark.unit
def test_is_fresh_ignores_expectation_ordering(in_memory_runtime_context):
    context = in_memory_runtime_context

    expectation_1 = gxe.ExpectColumnMeanToBeBetween(column="first")
    expectation_2 = gxe.ExpectColumnMeanToBeBetween(column="second")
    expectation_3 = gxe.ExpectColumnMeanToBeBetween(column="third")

    suite = context.suites.add(
        ExpectationSuite(
            name="my_suite", expectations=[expectation_1, expectation_2, expectation_3]
        )
    )

    suite_dict = {
        "name": "my_suite",
        # Order of expectations from the store is not the same as the original suite
        "expectations": [
            {"type": e.expectation_type, "kwargs": {"column": e.column}, "id": e.id}
            for e in (expectation_2, expectation_3, expectation_1)
        ],
    }
    with mock.patch.object(ExpectationsStore, "get", return_value=suite_dict):
        diagnostics = suite.is_fresh()

    assert diagnostics.success is True


class TestExpectationSuiteHash:
    @pytest.mark.unit
    def test_hash_consistency_with_equality(self, empty_data_context):
        expectation1 = ExpectColumnValuesToNotBeNull(column="test_column")
        expectation2 = ExpectColumnValuesToNotBeNull(column="test_column")

        suite1 = ExpectationSuite(name="test_suite")
        suite1.add_expectation(expectation1)

        suite2 = ExpectationSuite(name="test_suite")
        suite2.add_expectation(expectation2)

        assert suite1 == suite2
        assert hash(suite1) == hash(suite2)

    @pytest.mark.unit
    def test_hash_different_for_different_names(self, empty_data_context):
        expectation = ExpectColumnValuesToNotBeNull(column="test_column")

        suite1 = ExpectationSuite(name="test_suite_1")
        suite1.add_expectation(expectation)

        suite2 = ExpectationSuite(name="test_suite_2")
        suite2.add_expectation(expectation)

        assert suite1 != suite2
        assert hash(suite1) != hash(suite2)

    @pytest.mark.unit
    def test_hash_different_for_different_expectations(self, empty_data_context):
        expectation1 = ExpectColumnValuesToNotBeNull(column="test_column_1")
        expectation2 = ExpectColumnValuesToNotBeNull(column="test_column_2")

        suite1 = ExpectationSuite(name="test_suite")
        suite1.add_expectation(expectation1)

        suite2 = ExpectationSuite(name="test_suite")
        suite2.add_expectation(expectation2)

        assert suite1 != suite2
        assert hash(suite1) != hash(suite2)

    @pytest.mark.unit
    def test_hash_different_for_different_meta(self, empty_data_context):
        expectation = ExpectColumnValuesToNotBeNull(column="test_column")

        suite1 = ExpectationSuite(name="test_suite", meta={"test": "value1"})
        suite1.add_expectation(expectation)

        suite2 = ExpectationSuite(name="test_suite", meta={"test": "value2"})
        suite2.add_expectation(expectation)

        assert suite1 != suite2
        assert hash(suite1) != hash(suite2)

    @pytest.mark.unit
    def test_hash_different_for_different_suite_parameters(self, empty_data_context):
        expectation = ExpectColumnValuesToNotBeNull(column="test_column")

        suite1 = ExpectationSuite(name="test_suite", suite_parameters={"param1": "value1"})
        suite1.add_expectation(expectation)

        suite2 = ExpectationSuite(name="test_suite", suite_parameters={"param1": "value2"})
        suite2.add_expectation(expectation)

        assert suite1 != suite2
        assert hash(suite1) != hash(suite2)

    @pytest.mark.unit
    def test_hash_stable_across_runs(self, empty_data_context):
        expectation = ExpectColumnValuesToNotBeNull(column="test_column")
        suite = ExpectationSuite(name="test_suite")
        suite.add_expectation(expectation)

        hash1 = hash(suite)
        hash2 = hash(suite)
        hash3 = hash(suite)

        assert hash1 == hash2 == hash3



from __future__ import annotations

import logging
import pathlib
from pprint import pformat as pf
from typing import Final

import pytest
import tomli
from tasks import MARKER_DEPENDENCY_MAP

pytestmark = pytest.mark.project

LOGGER: Final = logging.getLogger(__name__)
PROJECT_ROOT: Final = pathlib.Path(__file__).parent.parent
PYPROJECT_TOML: Final = PROJECT_ROOT / "pyproject.toml"
# Markers that are used to launch CI but map to a different marker for tests.
# eg, gx-redshift should run the redshift test so, while a marker for CI
# there should be no tests with this marker.
NO_TEST_MARKERS: Final = ["gx-redshift"]


@pytest.fixture(scope="module")
def pyproject_toml_dict() -> dict:
    """Parse pyporject.toml and return as dict"""
    return tomli.loads(PYPROJECT_TOML.read_text())


@pytest.fixture(scope="module")
def pytest_markers(pyproject_toml_dict: dict) -> list[str]:
    """Return pytest markers"""
    LOGGER.debug(f"pytest config ->\n{pf(pyproject_toml_dict['tool']['pytest'], depth=2)}")
    marker_details = pyproject_toml_dict["tool"]["pytest"]["ini_options"]["markers"]
    LOGGER.debug(f"marker_details ->\n{pf(marker_details)}")
    return [m.split(":")[0] for m in marker_details]


def test_marker_mappings_are_registered(pytest_markers: list[str]):
    """
    Check that all pytest marker mappings are actually valid,
    and have been registered with pytest.
    """
    LOGGER.debug(f"pytest_markers:\n----------\n{pf(pytest_markers)}")

    for marker in MARKER_DEPENDENCY_MAP:
        if marker in NO_TEST_MARKERS:
            continue
        assert marker in pytest_markers


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



from __future__ import annotations

import re
from typing import List, Literal, Tuple, Union

from great_expectations.compatibility import pydantic


class ColumnPartitionerYearly(pydantic.BaseModel):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year"] = "partition_on_year"


class ColumnPartitionerMonthly(pydantic.BaseModel):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year_and_month"] = "partition_on_year_and_month"


class ColumnPartitionerDaily(pydantic.BaseModel):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year_and_month_and_day"] = (
        "partition_on_year_and_month_and_day"
    )


class PartitionerDatetimePart(pydantic.BaseModel):
    datetime_parts: List[str]
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_date_parts"] = "partition_on_date_parts"


class PartitionerDividedInteger(pydantic.BaseModel):
    divisor: int
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_divided_integer"] = "partition_on_divided_integer"


class PartitionerModInteger(pydantic.BaseModel):
    mod: int
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_mod_integer"] = "partition_on_mod_integer"


class PartitionerColumnValue(pydantic.BaseModel):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_column_value"] = "partition_on_column_value"


class PartitionerMultiColumnValue(pydantic.BaseModel):
    column_names: List[str]
    sort_ascending: bool = True
    method_name: Literal["partition_on_multi_column_values"] = "partition_on_multi_column_values"


class PartitionerConvertedDatetime(pydantic.BaseModel):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_converted_datetime"] = "partition_on_converted_datetime"
    date_format_string: str


ColumnPartitioner = Union[
    PartitionerColumnValue,
    PartitionerMultiColumnValue,
    PartitionerDividedInteger,
    PartitionerModInteger,
    ColumnPartitionerYearly,
    ColumnPartitionerMonthly,
    ColumnPartitionerDaily,
    PartitionerDatetimePart,
    PartitionerConvertedDatetime,
]


class FileNamePartitionerYearly(pydantic.BaseModel):
    regex: re.Pattern
    param_names: Tuple[Literal["year"]] = ("year",)
    sort_ascending: bool = True


class FileNamePartitionerMonthly(pydantic.BaseModel):
    regex: re.Pattern
    param_names: Tuple[Literal["year"], Literal["month"]] = ("year", "month")
    sort_ascending: bool = True


class FileNamePartitionerDaily(pydantic.BaseModel):
    regex: re.Pattern
    param_names: Tuple[Literal["year"], Literal["month"], Literal["day"]] = ("year", "month", "day")
    sort_ascending: bool = True


class FileNamePartitionerPath(pydantic.BaseModel):
    regex: re.Pattern
    param_names: Tuple[()] = ()
    sort_ascending: bool = True


FileNamePartitioner = Union[
    FileNamePartitionerYearly,
    FileNamePartitionerMonthly,
    FileNamePartitionerDaily,
    FileNamePartitionerPath,
]



import datetime

import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.batch_spec import RuntimeDataBatchSpec

# module level markers
pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "underscore_prefix",
    [
        pytest.param("_", id="underscore prefix"),
        pytest.param("", id="no underscore prefix"),
    ],
)
@pytest.mark.parametrize(
    "sampling_kwargs,num_sampled_rows",
    [
        pytest.param({"n": 3}, 3, id="sample n=3"),
        pytest.param({"n": 1000}, 120, id="sample n=1000 larger than df"),
        pytest.param({"n": 0}, 0, id="sample n=0"),
        pytest.param(
            {},
            0,
            id="n missing from sampling_kwargs",
            marks=pytest.mark.xfail(strict=True, raises=gx_exceptions.SamplerError),
        ),
        pytest.param(
            None,
            0,
            id="sampling_kwargs are None",
            marks=pytest.mark.xfail(strict=True, raises=gx_exceptions.SamplerError),
        ),
    ],
)
def test_limit_sampler_get_batch_data(
    sampling_kwargs,
    num_sampled_rows,
    underscore_prefix,
    test_sparkdf,
    basic_spark_df_execution_engine,
):
    sampled_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            sampling_method=f"{underscore_prefix}sample_using_limit",
            sampling_kwargs=sampling_kwargs,
        )
    ).dataframe

    assert sampled_df.count() == num_sampled_rows


def test_get_batch_empty_sampler(test_sparkdf, basic_spark_df_execution_engine):
    sampled_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(batch_data=test_sparkdf, sampling_method=None)
    ).dataframe
    assert sampled_df.count() == 120
    assert len(sampled_df.columns) == 10


def test_sample_using_random(test_sparkdf, basic_spark_df_execution_engine):
    sampled_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(batch_data=test_sparkdf, sampling_method="_sample_using_random")
    ).dataframe
    # The test dataframe contains 10 columns and 120 rows.
    assert len(sampled_df.columns) == 10
    assert 0 <= sampled_df.count() <= 120
    # The sampling probability "p" used in "SparkDFExecutionEngine._sample_using_random()" is 0.1 (the equivalent of an  # noqa: E501 # FIXME CoP
    # unfair coin with the 10% chance of coming up as "heads").  Hence, we should never get as much as 20% of the rows.  # noqa: E501 # FIXME CoP
    assert sampled_df.count() < 25


def test_sample_using_mod(test_sparkdf, basic_spark_df_execution_engine):
    sampled_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            sampling_method="_sample_using_mod",
            sampling_kwargs={
                "column_name": "id",
                "mod": 5,
                "value": 4,
            },
        )
    ).dataframe
    assert sampled_df.count() == 24
    assert len(sampled_df.columns) == 10


def test_sample_using_a_list(test_sparkdf, basic_spark_df_execution_engine):
    sampled_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            sampling_method="_sample_using_a_list",
            sampling_kwargs={
                "column_name": "id",
                "value_list": [3, 5, 7, 11],
            },
        )
    ).dataframe
    assert sampled_df.count() == 4
    assert len(sampled_df.columns) == 10


def test_sample_using_md5_wrong_hash_function_name(test_sparkdf, basic_spark_df_execution_engine):
    with pytest.raises(gx_exceptions.ExecutionEngineError):
        # noinspection PyUnusedLocal
        _ = basic_spark_df_execution_engine.get_batch_data(
            RuntimeDataBatchSpec(
                batch_data=test_sparkdf,
                sampling_method="_sample_using_hash",
                sampling_kwargs={
                    "column_name": "date",
                    "hash_function_name": "I_wont_work",
                },
            )
        ).dataframe


def test_sample_using_md5(test_sparkdf, basic_spark_df_execution_engine):
    sampled_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            sampling_method="_sample_using_hash",
            sampling_kwargs={
                "column_name": "date",
                "hash_function_name": "md5",
            },
        )
    ).dataframe
    assert sampled_df.count() == 10
    assert len(sampled_df.columns) == 10

    collected = sampled_df.collect()
    for val in collected:
        assert val.date in [datetime.date(2020, 1, 15), datetime.date(2020, 1, 29)]



from __future__ import annotations

import logging
from typing import Set, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.data_context_variables import (
    DataContextVariableSchema,
)
from great_expectations.data_context.store.configuration_store import ConfigurationStore
from great_expectations.data_context.types.base import DataContextConfig

logger = logging.getLogger(__name__)


class DataContextStore(ConfigurationStore):
    """
    A DataContextStore manages persistence around DataContextConfigs.
    """

    _configuration_class = DataContextConfig

    cloud_exclude_field_names: Set[DataContextVariableSchema] = {
        DataContextVariableSchema.CHECKPOINT_STORE_NAME,
        DataContextVariableSchema.DATASOURCES,
        DataContextVariableSchema.EXPECTATIONS_STORE_NAME,
        DataContextVariableSchema.VALIDATIONS_STORE_NAME,
    }

    @override
    def serialize(self, value: DataContextConfig) -> Union[dict, str]:
        """
        Please see `ConfigurationStore.serialize` for more information.

        Note that GX Cloud utilizes a subset of the config; as such, an explicit
        step to remove unnecessary keys is a required part of the serialization process.

        Args:
            value: DataContextConfig to serialize utilizing the configured StoreBackend.

        Returns:
            Either a string or dictionary representation of the serialized config.
        """
        payload: Union[str, dict] = super().serialize(value=value)

        # Cloud requires a subset of the DataContextConfig
        if self.cloud_mode:
            assert isinstance(payload, dict)
            for attr in self.cloud_exclude_field_names:
                if attr in payload:
                    payload.pop(attr)
                    logger.debug(f"Removed {attr} from DataContextConfig while serializing to JSON")

        return payload



from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable
from unittest import mock

import pandas as pd
import pytest

import great_expectations as gx
import great_expectations.expectations as gxe
from great_expectations.compatibility import pydantic
from great_expectations.core.partitioners import (
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
    PartitionerColumnValue,
    PartitionerConvertedDatetime,
    PartitionerDatetimePart,
    PartitionerDividedInteger,
    PartitionerModInteger,
    PartitionerMultiColumnValue,
)
from great_expectations.data_context import (
    AbstractDataContext,
    CloudDataContext,
    EphemeralDataContext,
    FileDataContext,
)
from great_expectations.datasource.fluent import (
    BatchRequest,
)
from great_expectations.datasource.fluent.interfaces import (
    DataAsset,
    Datasource,
)
from great_expectations.execution_engine.pandas_batch_data import PandasBatchData
from great_expectations.execution_engine.sparkdf_batch_data import SparkDFBatchData
from great_expectations.validator.v1_validator import Validator
from tests.integration.fluent.conftest import sqlite_datasource
from tests.integration.fluent.integration_test_utils import (
    run_batch_head,
    run_checkpoint_and_data_doc,
)

if TYPE_CHECKING:
    from responses import RequestsMock

    from great_expectations.compatibility.pyspark import DataFrame as SparkDataFrame
    from great_expectations.compatibility.pyspark import SparkSession
    from great_expectations.datasource.fluent.pandas_datasource import (
        DataFrameAsset as PandasDataFrameAsset,
    )
    from great_expectations.datasource.fluent.pandas_datasource import PandasDatasource
    from great_expectations.datasource.fluent.spark_datasource import (
        DataFrameAsset as SparkDataFrameAsset,
    )
    from great_expectations.datasource.fluent.spark_datasource import SparkDatasource


# This is marked by the various backend used in testing in the datasource_test_data fixture.
def test_run_checkpoint_and_data_doc(
    datasource_test_data: tuple[AbstractDataContext, Datasource, DataAsset, BatchRequest],
):
    run_checkpoint_and_data_doc(
        datasource_test_data=datasource_test_data,
    )


# This is marked by the various backend used in testing in the datasource_test_data fixture.
@pytest.mark.parametrize(
    ["n_rows", "fetch_all", "success"],
    [
        (None, False, True),
        (3, False, True),
        (7, False, True),
        (-100, False, True),
        ("invalid_value", False, False),
        (1.5, False, False),
        (True, False, False),
        (0, False, True),
        (200000, False, True),
        (1, False, True),
        (-50000, False, True),
        (-5, True, True),
        (0, True, True),
        (3, True, True),
        (50000, True, True),
        (-20000, True, True),
        (None, True, True),
        (15, "invalid_value", False),
    ],
)
def test_batch_head(
    datasource_test_data: tuple[AbstractDataContext, Datasource, DataAsset, BatchRequest],
    fetch_all: bool | str,
    n_rows: int | float | str | None,  # noqa: PYI041 # FIXME CoP
    success: bool,
) -> None:
    run_batch_head(
        datasource_test_data=datasource_test_data,
        fetch_all=fetch_all,
        n_rows=n_rows,
        success=success,
    )


@pytest.mark.sqlite
class TestQueryAssets:
    def test_success_with_partitioners(self, empty_data_context):
        context = empty_data_context
        datasource = sqlite_datasource(context, "yellow_tripdata.db")
        passenger_count_value = 5
        asset = datasource.add_query_asset(
            name="query_asset",
            query=f"   SELECT * from yellow_tripdata_sample_2019_02 WHERE passenger_count = {passenger_count_value}",  # noqa: E501 # FIXME CoP
        )
        validator = context.get_validator(
            batch_request=asset.build_batch_request(
                options={"year": 2019},
                partitioner=ColumnPartitionerMonthly(column_name="pickup_datetime"),
            )
        )
        result = validator.expect_column_distinct_values_to_equal_set(
            column="passenger_count",
            value_set=[passenger_count_value],
            result_format={"result_format": "BOOLEAN_ONLY"},
        )
        assert result.success

    def test_partitioner_filtering(self, empty_data_context):
        context = empty_data_context
        datasource = sqlite_datasource(context, "../../test_cases_for_sql_data_connector.db")

        asset = datasource.add_query_asset(
            name="trip_asset_partition_by_event_type",
            query="SELECT * FROM table_partitioned_by_date_column__A",
        )
        batch_request = asset.build_batch_request(
            options={"event_type": "start"},
            partitioner=PartitionerColumnValue(column_name="event_type"),
        )
        validator = context.get_validator(batch_request=batch_request)

        # All rows returned by head have the start event_type.
        result = validator.execution_engine.batch_manager.active_batch.head(n_rows=50)
        unique_event_types = set(result.data["event_type"].unique())
        print(f"{unique_event_types=}")
        assert unique_event_types == {"start"}


@pytest.mark.sqlite
@pytest.mark.parametrize(
    [
        "database",
        "table_name",
        "partitioner_class",
        "partitioner_kwargs",
        "all_batches_cnt",
        "specified_batch_request",
        "specified_batch_cnt",
        "last_specified_batch_metadata",
    ],
    [
        pytest.param(
            "yellow_tripdata_sample_2020_all_months_combined.db",
            "yellow_tripdata_sample_2020",
            ColumnPartitionerYearly,
            {"column_name": "pickup_datetime"},
            1,
            {"year": 2020},
            1,
            {"year": 2020},
            id="year",
        ),
        pytest.param(
            "yellow_tripdata_sample_2020_all_months_combined.db",
            "yellow_tripdata_sample_2020",
            ColumnPartitionerMonthly,
            {"column_name": "pickup_datetime"},
            12,
            {"year": 2020, "month": 6},
            1,
            {"year": 2020, "month": 6},
            id="year_and_month",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            ColumnPartitionerDaily,
            {"column_name": "pickup_datetime"},
            28,
            {"year": 2019, "month": 2, "day": 10},
            1,
            {"year": 2019, "month": 2, "day": 10},
            id="year_and_month_and_day",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            PartitionerDatetimePart,
            {
                "column_name": "pickup_datetime",
                "datetime_parts": ["year", "month", "day"],
            },
            28,
            {"year": 2019, "month": 2},
            28,
            {"year": 2019, "month": 2, "day": 28},
            id="datetime_part",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            PartitionerColumnValue,
            {"column_name": "passenger_count"},
            7,
            {"passenger_count": 3},
            1,
            {"passenger_count": 3},
            id="column_value",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            PartitionerColumnValue,
            {"column_name": "pickup_datetime"},
            9977,
            {"pickup_datetime": "2019-02-07 15:48:06"},
            1,
            {"pickup_datetime": "2019-02-07 15:48:06"},
            id="column_value_datetime",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            PartitionerDividedInteger,
            {"column_name": "passenger_count", "divisor": 3},
            3,
            {"quotient": 2},
            1,
            {"quotient": 2},
            id="divisor",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            PartitionerModInteger,
            {"column_name": "passenger_count", "mod": 3},
            3,
            {"remainder": 2},
            1,
            {"remainder": 2},
            id="mod_integer",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            PartitionerConvertedDatetime,
            {"column_name": "pickup_datetime", "date_format_string": "%Y-%m-%d"},
            28,
            {"datetime": "2019-02-23"},
            1,
            {"datetime": "2019-02-23"},
            id="converted_datetime",
        ),
        pytest.param(
            "yellow_tripdata.db",
            "yellow_tripdata_sample_2019_02",
            PartitionerMultiColumnValue,
            {"column_names": ["passenger_count", "payment_type"]},
            23,
            {"passenger_count": 1, "payment_type": 1},
            1,
            {"passenger_count": 1, "payment_type": 1},
            id="multi_column_values",
        ),
    ],
)
def test_partitioner(
    empty_data_context,
    database,
    table_name,
    partitioner_class,
    partitioner_kwargs,
    all_batches_cnt,
    specified_batch_request,
    specified_batch_cnt,
    last_specified_batch_metadata,
):
    context = empty_data_context
    datasource = sqlite_datasource(context, database)
    asset = datasource.add_table_asset(
        name="table_asset",
        table_name=table_name,
    )
    partitioner = partitioner_class(**partitioner_kwargs)
    # Test getting all batches
    all_batches = asset.get_batch_identifiers_list(
        asset.build_batch_request(partitioner=partitioner)
    )
    assert len(all_batches) == all_batches_cnt
    # Test getting specified batches
    specified_batches = asset.get_batch_identifiers_list(
        asset.build_batch_request(specified_batch_request, partitioner=partitioner)
    )
    assert len(specified_batches) == specified_batch_cnt
    assert specified_batches[-1] == last_specified_batch_metadata


@pytest.mark.sqlite
def test_partitioner_build_batch_request_allows_selecting_by_date_and_datetime_as_string(
    empty_data_context,
):
    context = empty_data_context
    datasource = sqlite_datasource(context, "yellow_tripdata.db")

    asset = datasource.add_query_asset(
        "query_asset",
        "SELECT date(pickup_datetime) as pickup_date, passenger_count FROM yellow_tripdata_sample_2019_02",  # noqa: E501 # FIXME CoP
    )
    partitioner = PartitionerColumnValue(column_name="pickup_date")
    # Test getting all batches
    all_batches = asset.get_batch_identifiers_list(
        asset.build_batch_request(partitioner=partitioner)
    )
    assert len(all_batches) == 28

    with mock.patch(
        "great_expectations.datasource.fluent.sql_datasource._partitioner_and_sql_asset_to_batch_identifier_data"
    ) as mock_batch_identifiers:
        mock_batch_identifiers.return_value = [
            {"pickup_date": datetime.date(2019, 2, 1)},
            {"pickup_date": datetime.date(2019, 2, 2)},
        ]
        specified_batches = asset.get_batch_identifiers_list(
            asset.build_batch_request(
                options={"pickup_date": "2019-02-01"}, partitioner=partitioner
            )
        )
        assert len(specified_batches) == 1

    with mock.patch(
        "great_expectations.datasource.fluent.sql_datasource._partitioner_and_sql_asset_to_batch_identifier_data"
    ) as mock_batch_identifiers:
        mock_batch_identifiers.return_value = [
            {"pickup_date": datetime.datetime(2019, 2, 1)},  # noqa: DTZ001 # FIXME CoP
            {"pickup_date": datetime.datetime(2019, 2, 2)},  # noqa: DTZ001 # FIXME CoP
        ]
        specified_batches = asset.get_batch_identifiers_list(
            asset.build_batch_request(
                options={"pickup_date": "2019-02-01 00:00:00"}, partitioner=partitioner
            )
        )
        assert len(specified_batches) == 1


@pytest.mark.parametrize(
    ["month", "expected"],
    [
        (1, 364),
        (2, 342),
    ],
)
@pytest.mark.sqlite
def test_success_with_partitioners_from_batch_definitions(
    empty_data_context,
    month: int,
    expected: int,
):
    """Integration test to ensure partitions from batch configs are used.

    The test is parameterized just to ensure that the partitioner is actually doing something.
    """
    context = empty_data_context
    datasource = sqlite_datasource(context, "yellow_tripdata_sample_2020_all_months_combined.db")
    passenger_count_value = 5
    asset = datasource.add_query_asset(
        name="query_asset",
        query=f"SELECT * from yellow_tripdata_sample_2020 WHERE passenger_count = {passenger_count_value}",  # noqa: E501 # FIXME CoP
    )
    batch_definition = asset.add_batch_definition(
        name="whatevs",
        partitioner=ColumnPartitionerMonthly(column_name="pickup_datetime"),
    )
    validator = Validator(
        batch_definition=batch_definition,
        batch_parameters={"year": 2020, "month": month},
    )
    result = validator.validate_expectation(gxe.ExpectTableRowCountToEqual(value=expected))
    assert result.success


@pytest.mark.parametrize(
    ["add_asset_method", "add_asset_kwarg"],
    [
        pytest.param(
            "add_table_asset",
            {"table_name": "yellow_tripdata_sample_2019_02"},
            id="table_asset",
        ),
        pytest.param(
            "add_query_asset",
            {"query": "select * from yellow_tripdata_sample_2019_02"},
            id="query_asset",
        ),
    ],
)
@pytest.mark.sqlite
def test_asset_specified_metadata(empty_data_context, add_asset_method, add_asset_kwarg):
    context = empty_data_context
    datasource = sqlite_datasource(context, "yellow_tripdata.db")
    asset_specified_metadata = {"pipeline_name": "my_pipeline"}
    asset = getattr(datasource, add_asset_method)(
        name="asset",
        batch_metadata=asset_specified_metadata,
        **add_asset_kwarg,
    )
    partitioner = ColumnPartitionerMonthly(column_name="pickup_datetime")
    # Test getting all batches
    batch = asset.get_batch(asset.build_batch_request(partitioner=partitioner))
    # Update the batch_metadata from the request with the metadata inherited from the asset
    assert batch.metadata == {**asset_specified_metadata, "year": 2019, "month": 2}


# This is marked by the various backend used in testing in the datasource_test_data fixture.
def test_batch_request_error_messages(
    datasource_test_data: tuple[AbstractDataContext, Datasource, DataAsset, BatchRequest],
) -> None:
    _, _, _, batch_request = datasource_test_data
    # DataAsset.build_batch_request() infers datasource_name and data_asset_name
    # which have already been confirmed as functional via test_connection() methods.
    with pytest.raises(TypeError):
        batch_request.datasource_name = "untested_datasource_name"

    with pytest.raises(TypeError):
        batch_request.data_asset_name = "untested_data_asset_name"

    # options can be added/updated if they take the correct form
    batch_request.options["new_option"] = 42
    assert "new_option" in batch_request.options

    with pytest.raises(pydantic.ValidationError):
        batch_request.options = {10: "value for non-string key"}  # type: ignore[dict-item] # FIXME CoP

    with pytest.raises(pydantic.ValidationError):
        batch_request.options = "not a dictionary"  # type: ignore[assignment] # FIXME CoP

    # batch_slice can be updated if it takes the correct form
    batch_request.batch_slice = "[5:10]"  # type: ignore[assignment] # FIXME CoP
    assert batch_request.batch_slice == slice(5, 10, None)

    # batch_slice can be updated via update method
    batch_request.update_batch_slice("[2:10:2]")
    assert batch_request.batch_slice == slice(2, 10, 2)

    with pytest.raises(ValueError):
        batch_request.batch_slice = "nonsense slice"  # type: ignore[assignment] # FIXME CoP

    with pytest.raises(ValueError):
        batch_request.batch_slice = True  # type: ignore[assignment] # FIXME CoP


@pytest.mark.cloud
def test_pandas_data_adding_dataframe_in_cloud_context(
    cloud_api_fake: RequestsMock,
    empty_cloud_context_fluent: CloudDataContext,
):
    df = pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})

    context = empty_cloud_context_fluent

    dataframe_asset: PandasDataFrameAsset = context.data_sources.add_or_update_pandas(
        name="fluent_pandas_datasource"
    ).add_dataframe_asset(name="my_df_asset")
    batch_def = dataframe_asset.add_batch_definition_whole_dataframe(name="bd")
    batch = batch_def.get_batch(batch_parameters={"dataframe": df})
    assert isinstance(batch.data, PandasBatchData)
    assert batch.data.dataframe.equals(df)


@pytest.mark.filesystem
def test_pandas_data_adding_dataframe_in_file_reloaded_context(
    empty_file_context: FileDataContext,
):
    df = pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})

    context = empty_file_context

    datasource = context.data_sources.add_or_update_pandas(name="fluent_pandas_datasource")
    dataframe_asset: PandasDataFrameAsset = datasource.add_dataframe_asset(name="my_df_asset")
    batch_def = dataframe_asset.add_batch_definition_whole_dataframe(name="bd")
    batch = batch_def.get_batch(batch_parameters={"dataframe": df})
    assert isinstance(batch.data, PandasBatchData)
    assert batch.data.dataframe.equals(df)

    # Reload the asset and see that we can re-add the df to the batch definition
    context = gx.get_context(context_root_dir=context.root_directory, cloud_mode=False)
    dataframe_asset = context.data_sources.get(name="fluent_pandas_datasource").get_asset(
        name="my_df_asset"
    )
    reloaded_batch_def = dataframe_asset.get_batch_definition(name="bd")
    batch = reloaded_batch_def.get_batch(batch_parameters={"dataframe": df})
    assert isinstance(batch.data, PandasBatchData)
    assert batch.data.dataframe.equals(df)


@pytest.mark.spark
def test_spark_data_adding_dataframe_in_cloud_context(
    spark_session,
    spark_df_from_pandas_df,
    cloud_api_fake: RequestsMock,
    empty_cloud_context_fluent: CloudDataContext,
):
    df = pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})
    spark_df = spark_df_from_pandas_df(spark_session, df)

    context = empty_cloud_context_fluent

    dataframe_asset: SparkDataFrameAsset = context.data_sources.add_or_update_spark(
        name="fluent_spark_datasource"
    ).add_dataframe_asset(name="my_df_asset")
    batch_def = dataframe_asset.add_batch_definition_whole_dataframe(name="bd")
    batch = batch_def.get_batch(batch_parameters={"dataframe": spark_df})
    assert isinstance(batch.data, SparkDFBatchData)
    assert batch.data.dataframe.toPandas().equals(df)


@pytest.mark.spark
def test_spark_data_adding_dataframe_in_file_reloaded_context(
    spark_session,
    spark_df_from_pandas_df,
    empty_file_context: FileDataContext,
):
    df = pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})
    spark_df = spark_df_from_pandas_df(spark_session, df)

    context = empty_file_context

    dataframe_asset: SparkDataFrameAsset = context.data_sources.add_or_update_spark(
        name="fluent_spark_datasource"
    ).add_dataframe_asset(name="my_df_asset")
    batch_def = dataframe_asset.add_batch_definition_whole_dataframe(name="bd")
    batch = batch_def.get_batch(batch_parameters={"dataframe": spark_df})
    assert isinstance(batch.data, SparkDFBatchData)
    assert batch.data.dataframe.toPandas().equals(df)

    context = gx.get_context(context_root_dir=context.root_directory, cloud_mode=False)
    retrieved_bd = (
        context.data_sources.get(name="fluent_spark_datasource")
        .get_asset(name="my_df_asset")
        .get_batch_definition(name="bd")
    )
    new_batch = retrieved_bd.get_batch(batch_parameters={"dataframe": spark_df})
    assert isinstance(new_batch.data, SparkDFBatchData)
    assert new_batch.data.dataframe.toPandas().equals(df)


@dataclass
class PandasDataSourceAndFrame:
    datasource: PandasDatasource
    dataframe: pd.DataFrame


@dataclass
class SparkDataSourceAndFrame:
    datasource: SparkDatasource
    dataframe: SparkDataFrame


def _validate_whole_dataframe_batch(
    source_and_frame: PandasDataSourceAndFrame | SparkDataSourceAndFrame,
):
    my_expectation = gxe.ExpectColumnMeanToBeBetween(
        column="column_name", min_value=2.5, max_value=3.5
    )
    asset = source_and_frame.datasource.add_dataframe_asset(name="asset")
    bd = asset.add_batch_definition_whole_dataframe(name="bd")
    batch = bd.get_batch(batch_parameters={"dataframe": source_and_frame.dataframe})
    result = batch.validate(my_expectation)
    assert result.success


@pytest.mark.unit
def test_validate_pandas_batch():
    context = gx.get_context(mode="ephemeral")
    datasource = context.data_sources.add_pandas(name="ds")
    df = pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})
    _validate_whole_dataframe_batch(PandasDataSourceAndFrame(datasource=datasource, dataframe=df))


@pytest.mark.spark
def test_validate_spark_batch(
    spark_session: SparkSession,
    spark_df_from_pandas_df: Callable[[SparkSession, pd.DataFrame], SparkDataFrame],
):
    context = gx.get_context(mode="ephemeral")
    datasource = context.data_sources.add_spark(name="ds")
    spark_df = spark_df_from_pandas_df(
        spark_session, pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})
    )
    _validate_whole_dataframe_batch(
        SparkDataSourceAndFrame(datasource=datasource, dataframe=spark_df)
    )


@dataclass
class ContextPandasDataSourceAndFrame:
    context: EphemeralDataContext
    datasource: PandasDatasource
    dataframe: pd.DataFrame


@dataclass
class ContextSparkDataSourceAndFrame:
    context: EphemeralDataContext
    datasource: SparkDatasource
    dataframe: SparkDataFrame


def _validate_whole_dataframe_batch_definition(
    context_source_frame: ContextPandasDataSourceAndFrame | ContextSparkDataSourceAndFrame,
):
    asset = context_source_frame.datasource.add_dataframe_asset(name="asset")
    bd = asset.add_batch_definition_whole_dataframe(name="bd")
    suite = context_source_frame.context.suites.add(gx.ExpectationSuite(name="suite"))
    suite.add_expectation(
        gxe.ExpectColumnMeanToBeBetween(column="column_name", min_value=2.5, max_value=3.5)
    )
    validation_def = context_source_frame.context.validation_definitions.add(
        gx.ValidationDefinition(
            name="vd",
            data=bd,
            suite=suite,
        )
    )
    result = validation_def.run(batch_parameters={"dataframe": context_source_frame.dataframe})
    assert result.success


@pytest.mark.unit
def test_validate_pandas_batch_definition():
    context = gx.get_context(mode="ephemeral")
    datasource = context.data_sources.add_pandas(name="ds")
    df = pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})
    _validate_whole_dataframe_batch_definition(
        ContextPandasDataSourceAndFrame(context=context, datasource=datasource, dataframe=df)
    )


@pytest.mark.spark
def test_validate_spark_batch_definition(
    spark_session: SparkSession,
    spark_df_from_pandas_df: Callable[[SparkSession, pd.DataFrame], SparkDataFrame],
):
    context = gx.get_context(mode="ephemeral")
    datasource = context.data_sources.add_spark(name="ds")
    spark_df = spark_df_from_pandas_df(
        spark_session, pd.DataFrame({"column_name": [1, 2, 3, 4, 5]})
    )
    _validate_whole_dataframe_batch_definition(
        ContextSparkDataSourceAndFrame(context=context, datasource=datasource, dataframe=spark_df)
    )



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

POSTGRESQL_NOT_IMPORTED = NotImported(
    "postgresql connection components are not installed, please 'pip install psycopg2'"
)

try:
    import psycopg2  # noqa: F401 # FIXME CoP
    import sqlalchemy.dialects.postgresql as postgresqltypes
except ImportError:
    postgresqltypes = POSTGRESQL_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import TEXT
except (ImportError, AttributeError):
    TEXT = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import CHAR
except (ImportError, AttributeError):
    CHAR = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import INTEGER
except (ImportError, AttributeError):
    INTEGER = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import SMALLINT
except (ImportError, AttributeError):
    SMALLINT = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import BIGINT
except (ImportError, AttributeError):
    BIGINT = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import TIMESTAMP
except (ImportError, AttributeError):
    TIMESTAMP = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import DATE
except (ImportError, AttributeError):
    DATE = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
except (ImportError, AttributeError):
    DOUBLE_PRECISION = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import BOOLEAN
except (ImportError, AttributeError):
    BOOLEAN = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP

try:
    from sqlalchemy.dialects.postgresql import NUMERIC
except (ImportError, AttributeError):
    NUMERIC = POSTGRESQL_NOT_IMPORTED  # type: ignore[misc, assignment] # FIXME CoP


class POSTGRESQL_TYPES:
    """Namespace for PostgreSQL dialect types."""

    TEXT = TEXT
    CHAR = CHAR
    INTEGER = INTEGER
    SMALLINT = SMALLINT
    BIGINT = BIGINT
    TIMESTAMP = TIMESTAMP
    DATE = DATE
    DOUBLE_PRECISION = DOUBLE_PRECISION
    BOOLEAN = BOOLEAN
    NUMERIC = NUMERIC



from great_expectations.compatibility.not_imported import NotImported

PYPD_NOT_IMPORTED = NotImported(
    "PagerDuty dependencies are not installed, please 'pip install pypd'"
)

try:
    import pypd
except ImportError:
    pypd = PYPD_NOT_IMPORTED



from datetime import datetime

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    DATA_SOURCES_THAT_SUPPORT_DATE_COMPARISONS,
    JUST_PANDAS_DATA_SOURCES,
)

COL_NAME = "my_col"

ONES_AND_THREES = pd.DataFrame({COL_NAME: [1, 3]})


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=ONES_AND_THREES)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMaxToBeBetween(column=COL_NAME, min_value=1, max_value=4)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {"observed_value": 3}


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(column=COL_NAME),
            id="vacuous_success",
        ),
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(column=COL_NAME, min_value=1, max_value=4),
            id="min_and_max",
        ),
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(column=COL_NAME, min_value=1),
            id="just_min",
        ),
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(column=COL_NAME, max_value=4),
            id="just_max",
        ),
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(
                column=COL_NAME, min_value=1, max_value=4, strict_min=True, strict_max=True
            ),
            id="strict_min_and_max",
        ),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=pd.DataFrame({COL_NAME: [1, 2, None]})
)
def test_success(batch_for_datasource: Batch, expectation: gxe.ExpectColumnMaxToBeBetween) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=DATA_SOURCES_THAT_SUPPORT_DATE_COMPARISONS,
    data=pd.DataFrame({COL_NAME: [datetime(2024, 11, 19).date(), datetime(2024, 11, 22).date()]}),  # noqa: DTZ001 # FIXME CoP
)
def test_dates(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column=COL_NAME,
        min_value=datetime(2024, 11, 20).date(),  # noqa: DTZ001 # FIXME CoP
        max_value=datetime(2024, 11, 22).date(),  # noqa: DTZ001 # FIXME CoP
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=pd.DataFrame({COL_NAME: [1, 2, None]})
)
def test_ignores_nulls(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMaxToBeBetween(column=COL_NAME, min_value=1, max_value=3)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(column=COL_NAME, min_value=4),
            id="just_min_fail",
        ),
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(
                column=COL_NAME, min_value=3, strict_min=True, max_value=100
            ),
            id="strict_min_fail",
        ),
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(column=COL_NAME, max_value=2),
            id="just_max_fail",
        ),
        pytest.param(
            gxe.ExpectColumnMaxToBeBetween(
                column=COL_NAME, min_value=1, max_value=3, strict_max=True
            ),
            id="strict_max_fail",
        ),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ONES_AND_THREES
)
def test_failure(batch_for_datasource: Batch, expectation: gxe.ExpectColumnMaxToBeBetween) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=pd.DataFrame({COL_NAME: []})
)
def test_no_data(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column=COL_NAME, min_value=1, max_value=3, result_format=ResultFormat.SUMMARY
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success
    assert result.to_json_dict()["result"] == {"observed_value": None}


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(0, True, id="success"),
        pytest.param(4, False, id="failure"),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ONES_AND_THREES
)
def test_success_with_suite_param_min_value_(
    batch_for_datasource: Batch, suite_param_value: int, expected_result: bool
) -> None:
    suite_param_key = "expect_column_max_to_be_between"
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column=COL_NAME,
        min_value={"$PARAMETER": suite_param_key},
        max_value=5,
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(5, True, id="success"),
        pytest.param(2, False, id="failure"),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ONES_AND_THREES
)
def test_success_with_suite_param_max_value_(
    batch_for_datasource: Batch, suite_param_value: int, expected_result: bool
) -> None:
    suite_param_key = "expect_column_max_to_be_between"
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column=COL_NAME,
        min_value=0,
        max_value={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ONES_AND_THREES
)
def test_success_with_suite_param_strict_min_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "expect_column_max_to_be_between"
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column=COL_NAME,
        min_value=0,
        max_value=5,
        strict_min={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ONES_AND_THREES
)
def test_success_with_suite_param_strict_max_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "expect_column_max_to_be_between"
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column=COL_NAME,
        min_value=0,
        max_value=5,
        strict_max={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from __future__ import annotations

import pydantic

from great_expectations.compatibility.not_imported import (
    is_version_greater_or_equal,
)

if is_version_greater_or_equal(version=pydantic.VERSION, compare_version="2.0.0"):
    # TODO: don't use star imports
    from pydantic.v1 import *  # noqa: F403 # FIXME CoP
    from pydantic.v1 import (
        AnyUrl,
        BaseSettings,
        HttpUrl,
        StrictStr,
        UrlError,
        error_wrappers,
        errors,
        fields,
        generics,
        json,
        networks,
        schema,
        typing,
    )
    from pydantic.v1.generics import GenericModel
    from pydantic.v1.main import ModelMetaclass

else:
    # TODO: don't use star imports
    from pydantic import *  # type: ignore[assignment,no-redef] # noqa: F403 # FIXME CoP
    from pydantic import (  # type: ignore[no-redef] # FIXME CoP
        AnyUrl,
        BaseSettings,
        HttpUrl,
        StrictStr,
        UrlError,
        error_wrappers,
        errors,
        fields,
        generics,
        json,
        networks,
        schema,
        typing,
    )
    from pydantic.generics import GenericModel  # type: ignore[no-redef] # FIXME CoP
    from pydantic.main import ModelMetaclass  # type: ignore[no-redef] # FIXME CoP

__all__ = [
    "AnyUrl",
    "BaseSettings",
    "GenericModel",
    "HttpUrl",
    "ModelMetaclass",
    "StrictStr",
    "UrlError",
    "error_wrappers",
    "errors",
    "fields",
    "generics",
    "json",
    "networks",
    "schema",
    "typing",
]



from __future__ import annotations

import logging
import warnings

logger = logging.getLogger(__name__)


class MetaMetricProvider(type):
    """MetaMetricProvider registers metrics as they are defined."""

    def __new__(cls, clsname, bases, attrs):
        newclass = super().__new__(cls, clsname, bases, attrs)
        # noinspection PyUnresolvedReferences
        newclass._register_metric_functions()
        return newclass


# The following is based on "https://stackoverflow.com/questions/9008444/how-to-warn-about-class-name-deprecation".
class DeprecatedMetaMetricProvider(MetaMetricProvider):
    """
    Goals:
        Instantiation of a deprecated class should raise a warning;
        Subclassing of a deprecated class should raise a warning;
        Support isinstance and issubclass checks.
    """

    # TODO: <Alex>All logging/warning directives should be placed into a common module to be imported as needed.</Alex>  # noqa: E501 # FIXME CoP
    # deprecated-v0.13.12
    warnings.simplefilter("default", category=DeprecationWarning)

    # Arguments: True -- suppresses the warnings; False -- outputs the warnings (to stderr).
    logging.captureWarnings(False)

    def __new__(cls, name, bases, classdict, *args, **kwargs):
        alias = classdict.get("_DeprecatedMetaMetricProvider__alias")

        if alias is not None:

            def new(cls, *args, **kwargs):
                alias = cls._DeprecatedMetaMetricProvider__alias

                if alias is not None:
                    # deprecated-v0.13.12
                    warnings.warn(
                        f"""{cls.__name__} has been renamed to {alias} -- the alias {cls.__name__} is \
deprecated as of v0.13.12 and will be removed in v0.16.
""",  # noqa: E501 # FIXME CoP
                        DeprecationWarning,
                        stacklevel=2,
                    )

                return alias(*args, **kwargs)

            classdict["__new__"] = new
            classdict["_DeprecatedMetaMetricProvider__alias"] = alias

        fixed_bases = []

        for b in bases:
            alias = getattr(b, "_DeprecatedMetaMetricProvider__alias", None)

            if alias is not None:
                # deprecated-v0.13.12
                warnings.warn(
                    f"""{b.__name__} has been renamed to {alias.__name__} -- the alias {b.__name__} is deprecated \
as of v0.13.12 and will be removed in v0.16.
""",  # noqa: E501 # FIXME CoP
                    DeprecationWarning,
                    stacklevel=2,
                )

            # Avoid duplicate base classes.
            b = alias or b  # noqa: PLW2901 # FIXME CoP
            if b not in fixed_bases:
                fixed_bases.append(b)

        fixed_bases = tuple(fixed_bases)

        return super().__new__(cls, name, fixed_bases, classdict, *args, **kwargs)

    def __instancecheck__(cls, instance):  # type: ignore[explicit-override] # FIXME
        return any(cls.__subclasscheck__(c) for c in (type(instance), instance.__class__))

    def __subclasscheck__(cls, subclass):  # type: ignore[explicit-override] # FIXME
        if subclass is cls:
            return True
        else:
            return issubclass(subclass, cls._DeprecatedMetaMetricProvider__alias)



from __future__ import annotations

import pathlib
from typing import Tuple

import pytest

import great_expectations as gx
import great_expectations.expectations as gxe
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.interfaces import Batch, Datasource
from great_expectations.expectations.expectation import Expectation
from great_expectations.metrics import BatchRowCount, ColumnValuesNonNull, Metric
from great_expectations.metrics.batch.row_count import BatchRowCountResult
from great_expectations.metrics.column.values_non_null import ColumnValuesNonNullResult
from great_expectations.metrics.metric_results import MetricErrorResult

DATASOURCE_NAME = "my_pandas"
ASSET_NAME = "my_csv"


@pytest.fixture
def pandas_setup(csv_path: pathlib.Path) -> Tuple[AbstractDataContext, Batch]:
    context = gx.get_context(mode="ephemeral")
    source = context.data_sources.add_pandas(DATASOURCE_NAME)
    filepath = (
        csv_path
        / "ten_trips_from_each_month"
        / "yellow_tripdata_sample_10_trips_from_each_month.csv"
    )
    asset = source.add_csv_asset(ASSET_NAME, filepath_or_buffer=filepath)
    batch = asset.get_batch(asset.build_batch_request())
    return context, batch


@pytest.mark.filesystem
def test_batch_validate_expectation(pandas_setup: Tuple[AbstractDataContext, Batch]):
    _, batch = pandas_setup

    # Make Expectation
    expectation = gxe.ExpectColumnValuesToNotBeNull(
        column="vendor_id",
        mostly=0.95,
    )
    # Validate
    result = batch.validate(expectation)
    # Asserts on result
    assert result.success is True


@pytest.mark.filesystem
def test_batch_validate_expectation_suite(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    context, batch = pandas_setup

    # Make Expectation Suite
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    suite.add_expectation(
        gxe.ExpectColumnValuesToNotBeNull(
            column="vendor_id",
            mostly=0.95,
        )
    )
    # Validate
    result = batch.validate(suite)
    # Asserts on result
    assert result.success is True


@pytest.mark.filesystem
def test_batch_validate_expectation_with_expectation_params(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    _, batch = pandas_setup

    expectation = gx.expectations.ExpectColumnMaxToBeBetween(
        column="passenger_count",
        min_value={"$PARAMETER": "expect_passenger_max_to_be_above"},
        max_value={"$PARAMETER": "expect_passenger_max_to_be_below"},
    )
    result = batch.validate(
        expectation,
        expectation_parameters={
            "expect_passenger_max_to_be_above": 1,
            "expect_passenger_max_to_be_below": 10,
        },
    )
    # Asserts on result
    assert result.success is True


@pytest.mark.filesystem
def test_batch_validate_expectation_suite_with_expectation_params(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    context, batch = pandas_setup

    # Make Expectation Suite
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    suite.add_expectation(
        gx.expectations.ExpectColumnMaxToBeBetween(
            column="passenger_count",
            min_value={"$PARAMETER": "expect_passenger_max_to_be_above"},
            max_value={"$PARAMETER": "expect_passenger_max_to_be_below"},
        )
    )
    # Validate
    result = batch.validate(
        suite,
        expectation_parameters={
            "expect_passenger_max_to_be_above": 1,
            "expect_passenger_max_to_be_below": 10,
        },
    )
    # Asserts on result
    assert result.success is True


@pytest.mark.filesystem
def test_batch_validate_with_updated_expectation(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    _, batch = pandas_setup

    # Make Expectation
    expectation = gxe.ExpectColumnValuesToNotBeNull(
        column="vendor_id",
    )
    # Validate
    result = batch.validate(expectation)
    # Asserts on result
    assert result.success is False
    # Update expectation and validate
    expectation.mostly = 0.95
    result = batch.validate(expectation)
    assert result.success is True


@pytest.mark.filesystem
def test_batch_validate_expectation_suite_with_updated_expectation(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    context, batch = pandas_setup

    # Make Expectation Suite
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    suite.add_expectation(gxe.ExpectColumnValuesToNotBeNull(column="vendor_id"))
    # Validate
    result = batch.validate(suite)
    # Asserts on result
    assert result.success is False
    # Update suite and validate
    assert len(suite.expectations) == 1

    expectation = suite.expectations[0]
    assert isinstance(expectation, gxe.ExpectColumnValuesToNotBeNull)
    expectation.mostly = 0.95

    expectation.save()
    assert isinstance(suite.expectations[0], gxe.ExpectColumnValuesToNotBeNull)
    assert suite.expectations[0].mostly == 0.95

    result = batch.validate(suite)
    assert result.success is True


class TestBatchValidateExpectation:
    @pytest.fixture
    def expectation(self) -> Expectation:
        return gxe.ExpectColumnValuesToNotBeNull(column="vendor_id", mostly=0.95)

    @pytest.mark.filesystem
    def test_boolean_validation_result(
        self,
        pandas_setup: Tuple[AbstractDataContext, Batch],
        expectation: Expectation,
    ):
        _, batch = pandas_setup
        result = batch.validate(expectation, result_format="BOOLEAN_ONLY")

        assert result.success is True
        assert len(result.result) == 0

    @pytest.mark.filesystem
    def test_summary_validation_result(
        self,
        pandas_setup: Tuple[AbstractDataContext, Batch],
        expectation: Expectation,
    ):
        _, batch = pandas_setup
        summary_result = batch.validate(expectation, result_format="SUMMARY")

        assert summary_result.success is True
        assert len(summary_result.result) > 0

    @pytest.mark.filesystem
    def test_complete_validation_result(
        self,
        pandas_setup: Tuple[AbstractDataContext, Batch],
        expectation: Expectation,
    ):
        _, batch = pandas_setup
        result = batch.validate(expectation, result_format="COMPLETE")

        assert result.success is True
        assert "unexpected_index_list" in result.result


class TestBatchValidateExpectationSuite:
    @pytest.fixture
    def suite(self) -> ExpectationSuite:
        return gx.ExpectationSuite(
            name="my-suite",
            expectations=[gxe.ExpectColumnValuesToNotBeNull(column="vendor_id", mostly=0.95)],
        )

    @pytest.mark.filesystem
    def test_boolean_validation_result(
        self,
        pandas_setup: Tuple[AbstractDataContext, Batch],
        suite: ExpectationSuite,
    ):
        _, batch = pandas_setup
        result = batch.validate(suite, result_format="BOOLEAN_ONLY")

        assert result.success is True
        assert len(result.results[0].result) == 0

    @pytest.mark.filesystem
    def test_summary_validation_result(
        self,
        pandas_setup: Tuple[AbstractDataContext, Batch],
        suite: ExpectationSuite,
    ):
        _, batch = pandas_setup
        summary_result = batch.validate(suite, result_format="SUMMARY")

        assert summary_result.success is True
        assert len(summary_result.results[0].result) > 0

    @pytest.mark.filesystem
    def test_complete_validation_result(
        self,
        pandas_setup: Tuple[AbstractDataContext, Batch],
        suite: ExpectationSuite,
    ):
        _, batch = pandas_setup
        result = batch.validate(suite, result_format="COMPLETE")

        assert result.success is True
        assert "unexpected_index_list" in result.results[0].result


@pytest.mark.filesystem
def test_batch_validate_expectation_does_not_persist_a_batch_definition(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    context, batch = pandas_setup
    datasource = context.data_sources.get(DATASOURCE_NAME)
    assert isinstance(datasource, Datasource)
    asset = datasource.get_asset(ASSET_NAME)

    expectation = gxe.ExpectColumnValuesToNotBeNull(
        column="vendor_id",
        mostly=0.95,
    )
    result = batch.validate(expectation)

    assert result.success
    assert len(asset.batch_definitions) == 0


@pytest.mark.filesystem
def test_batch_validate_expectation_suite_does_not_persist_a_batch_definition(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    context, batch = pandas_setup
    datasource = context.data_sources.get(DATASOURCE_NAME)
    assert isinstance(datasource, Datasource)
    asset = datasource.get_asset(ASSET_NAME)

    suite = ExpectationSuite(
        "suite",
        expectations=[
            gxe.ExpectColumnValuesToNotBeNull(
                column="vendor_id",
                mostly=0.95,
            )
        ],
    )
    result = batch.validate(suite)

    assert result.success
    assert len(asset.batch_definitions) == 0


@pytest.mark.filesystem
def test_batch_compute_metrics_single_metric_success(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    _, batch = pandas_setup
    metric = ColumnValuesNonNull(
        column="vendor_id",
    )
    metric_results = batch.compute_metrics(metric)
    assert type(metric_results) is ColumnValuesNonNullResult


@pytest.mark.filesystem
def test_batch_compute_metrics_multiple_metrics_success(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    _, batch = pandas_setup
    metric_1 = ColumnValuesNonNull(
        column="passenger_count",
    )
    metric_2 = BatchRowCount()
    metrics: list[Metric] = [metric_1, metric_2]
    requested_metric_count = len(metrics)
    metric_results = batch.compute_metrics(metrics)
    assert isinstance(metric_results, list)
    assert len(metric_results) == requested_metric_count
    assert type(metric_results[0]) is ColumnValuesNonNullResult
    assert type(metric_results[1]) is BatchRowCountResult


@pytest.mark.filesystem
def test_batch_compute_metrics_multiple_metrics_error(
    pandas_setup: Tuple[AbstractDataContext, Batch],
):
    _, batch = pandas_setup
    metric_1 = ColumnValuesNonNull(
        column="not_a_column",
    )
    metric_2 = BatchRowCount()
    metrics: list[Metric] = [metric_1, metric_2]
    requested_metric_count = len(metrics)
    metric_results = batch.compute_metrics(metrics)
    assert isinstance(metric_results, list)
    assert len(metric_results) == requested_metric_count
    assert type(metric_results[0]) is MetricErrorResult
    assert type(metric_results[1]) is BatchRowCountResult



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Tuple, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import (
    COLUMN_DESCRIPTION,
    MOSTLY_DESCRIPTION,
)
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)

if TYPE_CHECKING:
    from great_expectations.render.renderer_configuration import AddParamArgs

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the Z-scores of a column's values to be less than a given threshold."
)
THRESHOLD_DESCRIPTION = (
    "A maximum Z-score threshold. "
    "All column Z-scores that are lower than this threshold will evaluate successfully."
)
DOUBLE_SIDED_DESCRIPTION = (
    "A True or False value indicating whether to evaluate double sidedly. Examples: "
    "(double_sided = True, threshold = 2) -> Z scores in non-inclusive interval(-2,2) | "
    "(double_sided = False, threshold = 2) -> Z scores in non-inclusive interval (-infinity,2)"
)
DATA_QUALITY_ISSUES = [DataQualityIssues.NUMERIC.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectColumnValueZScoresToBeLessThan(ColumnMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnValueZScoresToBeLessThan is a \
    Column Map Expectation \
    for typed-column backends, and also for PandasExecutionEngine where the column \
    dtype and provided type_ are unambiguous constraints \
    (any dtype except 'object' or dtype of 'object' with type_ specified as 'object').

    Column Map Expectations are one of the most common types of Expectation.
    They are evaluated for a single column and ask a yes/no question for every row in that column.
    Based on the result, they then calculate the percentage of rows that gave a positive answer. If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        threshold (number): \
            {THRESHOLD_DESCRIPTION}
        double_sided (boolean): \
            {DOUBLE_SIDED_DESCRIPTION}

    Other Parameters:
        mostly (None or a float between 0 and 1): \
            {MOSTLY_DESCRIPTION} \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	1       -100000000000
            1 	1       -1
            2 	1       0
            3   3       1
            4   3       1

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnValueZScoresToBeLessThan(
                    column="test",
                    threshold=1.96,
                    double_sided=True
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 5,
                    "unexpected_count": 0,
                    "unexpected_percent": 0.0,
                    "partial_unexpected_list": [],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 0.0,
                    "unexpected_percent_nonmissing": 0.0
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnValueZScoresToBeLessThan(
                    column="test2",
                    threshold=1,
                    double_sided=True
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 5,
                    "unexpected_count": 1,
                    "unexpected_percent": 20.0,
                    "partial_unexpected_list": [
                      -100000000000
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 20.0,
                    "unexpected_percent_nonmissing": 20.0
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    threshold: Union[float, SuiteParameterDict] = pydantic.Field(description=THRESHOLD_DESCRIPTION)
    double_sided: Union[bool, SuiteParameterDict] = pydantic.Field(
        description=DOUBLE_SIDED_DESCRIPTION
    )
    domain_keys: ClassVar[Tuple[str, ...]] = (
        "column",
        "row_condition",
        "condition_parser",
    )

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    # Setting necessary computation metric dependencies and defining kwargs, as well as assigning kwargs default values\  # noqa: E501 # FIXME CoP
    map_metric = "column_values.z_score.under_threshold"
    success_keys = ("threshold", "double_sided", "mostly")
    args_keys = ("column", "threshold")

    class Config:
        title = "Expect column value z-scores to be less than"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnValueZScoresToBeLessThan]
        ) -> None:
            ColumnMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @override
    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("threshold", RendererValueType.NUMBER),
            ("mostly", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if renderer_configuration.include_column_name:
            template_str = "$column value z-scores must be "
        else:
            template_str = "Value z-scores must be "

        if renderer_configuration.kwargs.get("double_sided") is True:
            inverse_threshold = params.threshold.value * -1
            renderer_configuration.add_param(
                name="inverse_threshold",
                param_type=RendererValueType.NUMBER,
                value=inverse_threshold,
            )
            if inverse_threshold < params.threshold.value:
                template_str += "greater than $inverse_threshold and less than $threshold"
            else:
                template_str += "greater than $threshold and less than $inverse_threshold"
        else:
            template_str += "less than $threshold"

        if params.mostly and params.mostly.value < 1.0:
            renderer_configuration = cls._add_mostly_pct_param(
                renderer_configuration=renderer_configuration
            )
            template_str += ", at least $mostly_pct % of the time."
        else:
            template_str += "."

        renderer_configuration.template_str = template_str

        return renderer_configuration



import uuid
from unittest import mock
from unittest.mock import Mock  # noqa: TID251 # FIXME CoP
from uuid import UUID

import numpy
import pytest

from great_expectations.data_context import CloudDataContext
from great_expectations.experimental.metric_repository.cloud_data_store import (
    CloudDataStore,
)
from great_expectations.experimental.metric_repository.metrics import (
    ColumnMetric,
    ColumnQuantileValuesMetric,
    MetricException,
    MetricRun,
)


@pytest.mark.cloud  # NOTE: needs orjson dependency
class TestCloudDataStoreMetricRun:
    def test_add_metric_run_non_generic_metric_type(
        self,
        empty_cloud_context_fluent: CloudDataContext,  # used as a fixture
    ):
        cloud_data_store = CloudDataStore(context=empty_cloud_context_fluent)
        data_asset_id = UUID("4469ed3b-61d4-421f-9635-8339d2558b0f")
        metric_run = MetricRun(
            data_asset_id=data_asset_id,
            metrics=[
                ColumnQuantileValuesMetric(
                    batch_id="batch_id",
                    metric_name="metric_name",
                    value=[0.25, 0.5, 0.75],
                    exception=None,
                    column="column",
                    quantiles=[0.25, 0.5, 0.75],
                    allow_relative_error=0.001,
                )
            ],
        )
        cloud_data_store._session = Mock()
        cloud_data_store._session.post = Mock()
        response_mock = Mock()
        cloud_data_store._session.post.return_value = response_mock

        response_metric_run_id = uuid.uuid4()
        response_metric_id = uuid.uuid4()
        response_mock.json.return_value = {
            "id": str(response_metric_run_id),
            "data_asset_id": str(data_asset_id),
            "metrics": [
                {
                    "id": str(response_metric_id),
                    "metric_type": "ColumnQuantileValuesMetric",
                    "value_type": "list[float]",
                    "allow_relative_error": 0.001,
                    "batch_id": "batch_id",
                    "column": "column",
                    "exception": None,
                    "metric_name": "metric_name",
                    "quantiles": [0.25, 0.5, 0.75],
                    "value": [0.25, 0.5, 0.75],
                }
            ],
        }

        uuid_from_add = cloud_data_store.add(metric_run)

        expected_data = '{"data":{"data_asset_id":"4469ed3b-61d4-421f-9635-8339d2558b0f","metrics":[{"batch_id":"batch_id","metric_name":"metric_name","value":[0.25,0.5,0.75],"exception":null,"column":"column","quantiles":[0.25,0.5,0.75],"allow_relative_error":0.001,"value_type":"list[float]","metric_type":"ColumnQuantileValuesMetric"}]}}'  # noqa: E501 # FIXME CoP

        cloud_data_store._session.post.assert_called_once_with(
            url="https://app.greatexpectations.fake.io/api/v1/organizations/12345678-1234-5678-1234-567812345678/metric-runs",
            data=expected_data,
        )
        assert uuid_from_add == response_metric_run_id

    def test_add_metric_run_generic_metric_type(
        self,
        empty_cloud_context_fluent: CloudDataContext,  # used as a fixture
    ):
        cloud_data_store = CloudDataStore(context=empty_cloud_context_fluent)
        data_asset_id = UUID("4469ed3b-61d4-421f-9635-8339d2558b0f")
        metric_run = MetricRun(
            data_asset_id=data_asset_id,
            metrics=[
                ColumnMetric[int](
                    batch_id="batch_id",
                    metric_name="metric_name",
                    value=1,
                    exception=None,
                    column="column",
                )
            ],
        )
        cloud_data_store._session = Mock()
        cloud_data_store._session.post = Mock()
        response_mock = Mock()
        cloud_data_store._session.post.return_value = response_mock

        response_metric_run_id = uuid.uuid4()
        response_metric_id = uuid.uuid4()
        response_mock.json.return_value = {
            "id": str(response_metric_run_id),
            "data_asset_id": str(data_asset_id),
            "metrics": [
                {
                    "id": str(response_metric_id),
                    "metric_type": "ColumnMetric",
                    "value_type": "int",
                    "batch_id": "batch_id",
                    "column": "column",
                    "exception": None,
                    "metric_name": "metric_name",
                    "value": 1,
                }
            ],
        }

        uuid_from_add = cloud_data_store.add(metric_run)

        expected_data = '{"data":{"data_asset_id":"4469ed3b-61d4-421f-9635-8339d2558b0f","metrics":[{"batch_id":"batch_id","metric_name":"metric_name","value":1,"exception":null,"column":"column","value_type":"int","metric_type":"ColumnMetric"}]}}'  # noqa: E501 # FIXME CoP

        cloud_data_store._session.post.assert_called_once_with(
            url="https://app.greatexpectations.fake.io/api/v1/organizations/12345678-1234-5678-1234-567812345678/metric-runs",
            data=expected_data,
        )
        assert uuid_from_add == response_metric_run_id

    def test_add_metric_run_generic_metric_type_with_exception(
        self,
        empty_cloud_context_fluent: CloudDataContext,  # used as a fixture
    ):
        cloud_data_store = CloudDataStore(context=empty_cloud_context_fluent)
        data_asset_id = UUID("4469ed3b-61d4-421f-9635-8339d2558b0f")
        metric_run = MetricRun(
            data_asset_id=data_asset_id,
            metrics=[
                ColumnMetric[int](
                    batch_id="batch_id",
                    metric_name="metric_name",
                    value=1,
                    exception=MetricException(type="exception type", message="exception message"),
                    column="column",
                )
            ],
        )
        cloud_data_store._session = Mock()
        cloud_data_store._session.post = Mock()
        response_mock = Mock()
        cloud_data_store._session.post.return_value = response_mock

        response_metric_run_id = uuid.uuid4()
        response_metric_id = uuid.uuid4()
        response_mock.json.return_value = {
            "id": str(response_metric_run_id),
            "data_asset_id": str(data_asset_id),
            "metrics": [
                {
                    "id": str(response_metric_id),
                    "metric_type": "ColumnMetric",
                    "value_type": "int",
                    "batch_id": "batch_id",
                    "column": "column",
                    "exception": {
                        "message": "exception message",
                        "type": "exception type",
                    },
                    "metric_name": "metric_name",
                    "value": 1,
                }
            ],
        }

        uuid_from_add = cloud_data_store.add(metric_run)

        expected_data = '{"data":{"data_asset_id":"4469ed3b-61d4-421f-9635-8339d2558b0f","metrics":[{"batch_id":"batch_id","metric_name":"metric_name","value":1,"exception":{"type":"exception type","message":"exception message"},"column":"column","value_type":"int","metric_type":"ColumnMetric"}]}}'  # noqa: E501 # FIXME CoP

        cloud_data_store._session.post.assert_called_once_with(
            url="https://app.greatexpectations.fake.io/api/v1/organizations/12345678-1234-5678-1234-567812345678/metric-runs",
            data=expected_data,
        )
        assert uuid_from_add == response_metric_run_id

    def test_add_metric_run_generic_metric_type_numpy(
        self,
        empty_cloud_context_fluent: CloudDataContext,  # used as a fixture
    ):
        cloud_data_store = CloudDataStore(context=empty_cloud_context_fluent)
        data_asset_id = UUID("4469ed3b-61d4-421f-9635-8339d2558b0f")
        metric_run = MetricRun(
            data_asset_id=data_asset_id,
            metrics=[
                ColumnMetric[numpy.float64](
                    batch_id="batch_id",
                    metric_name="metric_name",
                    value=numpy.float64(2.5),
                    exception=None,
                    column="column",
                )
            ],
        )
        cloud_data_store._session = Mock()
        cloud_data_store._session.post = Mock()
        response_mock = Mock()
        cloud_data_store._session.post.return_value = response_mock

        response_metric_run_id = uuid.uuid4()
        response_metric_id = uuid.uuid4()
        response_mock.json.return_value = {
            "id": str(response_metric_run_id),
            "data_asset_id": str(data_asset_id),
            "metrics": [
                {
                    "id": str(response_metric_id),
                    "metric_type": "ColumnMetric",
                    "value_type": "float64",
                    "batch_id": "batch_id",
                    "column": "column",
                    "exception": None,
                    "metric_name": "metric_name",
                    "value": 2.5,
                }
            ],
        }

        uuid_from_add = cloud_data_store.add(metric_run)

        expected_data = '{"data":{"data_asset_id":"4469ed3b-61d4-421f-9635-8339d2558b0f","metrics":[{"batch_id":"batch_id","metric_name":"metric_name","value":2.5,"exception":null,"column":"column","value_type":"float64","metric_type":"ColumnMetric"}]}}'  # noqa: E501 # FIXME CoP

        cloud_data_store._session.post.assert_called_once_with(
            url="https://app.greatexpectations.fake.io/api/v1/organizations/12345678-1234-5678-1234-567812345678/metric-runs",
            data=expected_data,
        )
        assert uuid_from_add == response_metric_run_id


@pytest.mark.cloud
def test_closes_session(
    empty_cloud_context_fluent: CloudDataContext,  # used as a fixture
):
    """Make sure that when the CloudDataStore object is garbage collected,
    the session is closed."""
    cloud_data_store = CloudDataStore(context=empty_cloud_context_fluent)
    with mock.patch("requests.Session.close", autospec=True) as mock_close:
        # Use the finalizer to remove the object from memory.
        # Using del or __del__ directly does not work in the test environment.
        cloud_data_store._finalizer()
        mock_close.assert_called_once()



from .meta_metric_provider import (  # isort:skip
    MetaMetricProvider,
    DeprecatedMetaMetricProvider,
)
from .column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
    ColumnMetricProvider,  # This class name is being deprecated (use "ColumnAggregateMetricProvider" going forward).  # noqa: E501 # FIXME CoP
    column_aggregate_partial,
    column_aggregate_value,
)
from .column_aggregate_metrics import *
from .column_map_metrics import *
from .column_pair_map_metrics import *
from .map_metric_provider import (
    ColumnMapMetricProvider,
    MapMetricProvider,
    column_condition_partial,
    column_function_partial,
)
from .multicolumn_map_metrics import *
from .query_metrics import *
from .table_metrics import *



from __future__ import annotations

import logging
import warnings
from typing import Callable, Iterator, Sequence

import pandas as pd

from great_expectations.compatibility import sqlalchemy
from great_expectations.compatibility.not_imported import is_version_less_than
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect

logger = logging.getLogger(__name__)


def read_sql_table_as_df(  # noqa: PLR0913 # FIXME CoP
    table_name,
    con,
    dialect: str,
    schema=None,
    index_col: str | Sequence[str] | None = None,
    coerce_float: bool = True,
    parse_dates: list[str] | dict[str, str] | None = None,
    columns: list[str] | None = None,
    chunksize: int | None = None,
) -> pd.DataFrame | Iterator[pd.DataFrame]:
    """Wrapper for `read_sql_table()` method in Pandas. Created as part of the effort to allow GX to be compatible
    with SqlAlchemy 2, and is used to suppress warnings that arise from implicit auto-commits.

    Args:
        table_name (str): name of SQL Table.
        con (sqlalchemy engine or connection): sqlalchemy.engine or sqlite3.Connection
        schema (str | None): Specify the schema (if database flavor supports this). If None, use
            default schema. Defaults to None.
        index_col (str | Sequence[str] | None): Column(s) to set as index(MultiIndex).
        coerce_float (bool): If True, method to convert values of non-string, non-numeric objects (like
            decimal.Decimal) to floating point. Can result in loss of Precision.
        parse_dates (List or Dict): list or dict, default None
            - List of column names to parse as dates.
            - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
            - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
        columns: List of column names to select from SQL table.
        chunksize: If specified, returns an iterator where `chunksize` is the number of
            rows to include in each chunk.
        dialect: we need to handle `sqlite` differently, so dialect is now optionally passed in.
    """  # noqa: E501 # FIXME CoP
    if is_version_less_than(pd.__version__, "2.0.0"):
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=DeprecationWarning)
            return _read_sql_table_as_df(
                table_name=table_name,
                con=con,
                dialect=dialect,
                schema=schema,
                index_col=index_col,
                coerce_float=coerce_float,
                parse_dates=parse_dates,
                columns=columns,
                chunksize=chunksize,
            )
    else:
        return _read_sql_table_as_df(
            table_name=table_name,
            con=con,
            dialect=dialect,
            schema=schema,
            index_col=index_col,
            coerce_float=coerce_float,
            parse_dates=parse_dates,
            columns=columns,
            chunksize=chunksize,
        )


def _read_sql_table_as_df(  # noqa: PLR0913 # FIXME CoP
    table_name,
    con,
    dialect: str,
    schema=None,
    index_col: str | Sequence[str] | None = None,
    coerce_float: bool = True,
    parse_dates: list[str] | dict[str, str] | None = None,
    columns: list[str] | None = None,
    chunksize: int | None = None,
) -> pd.DataFrame | Iterator[pd.DataFrame]:
    """Wrapper for `read_sql_table()` method in Pandas. Created as part of the effort to allow GX to be compatible
    with SqlAlchemy 2, and is used to suppress warnings that arise from implicit auto-commits.

    Args:
        table_name (str): name of SQL Table.
        con (sqlalchemy engine or connection): sqlalchemy.engine or sqlite3.Connection
        schema (str | None): Specify the schema (if database flavor supports this). If None, use
            default schema. Defaults to None.
        index_col (str | Sequence[str] | None): Column(s) to set as index(MultiIndex).
        coerce_float (bool): If True, method to convert values of non-string, non-numeric objects (like
            decimal.Decimal) to floating point. Can result in loss of Precision.
        parse_dates (List or Dict): list or dict, default None
            - List of column names to parse as dates.
            - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
            - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
        columns: List of column names to select from SQL table.
        chunksize: If specified, returns an iterator where `chunksize` is the number of
            rows to include in each chunk.
        dialect: we need to handle `sqlite` differently, so dialect is now optionally passed in.
    """  # noqa: E501 # FIXME CoP
    if dialect == GXSqlDialect.TRINO:
        return pd.read_sql_table(
            table_name=table_name,
            con=con,
            schema=schema,
            index_col=index_col,  # type: ignore[arg-type] # FIXME CoP
            coerce_float=coerce_float,
            parse_dates=parse_dates,
            columns=columns,
            chunksize=chunksize,  # type: ignore[arg-type] # FIXME CoP
        )
    else:
        sql_str: str
        if schema:
            sql_str = f"""SELECT * FROM {schema}.{table_name}"""
        else:
            sql_str = f"""SELECT * FROM {table_name}"""
        return pd.read_sql_query(
            sql=sql_str,
            con=con,
            index_col=index_col,  # type: ignore[arg-type] # FIXME CoP
            coerce_float=coerce_float,
            parse_dates=parse_dates,
            chunksize=chunksize,  # type: ignore[arg-type] # FIXME CoP
        )


def add_dataframe_to_db(  # noqa: PLR0913 # FIXME CoP
    df: pd.DataFrame,
    name: str,
    con,
    schema=None,
    if_exists: str = "fail",
    index: bool = True,
    index_label: str | None = None,
    chunksize: int | None = None,
    dtype: dict | None = None,
    method: str | Callable | None = None,
) -> None:
    """Write records stored in a DataFrame to a SQL database.

    Wrapper for `to_sql()` method in Pandas. Created as part of the effort to allow GX to be compatible
    with SqlAlchemy 2, and is used to suppress warnings that arise from implicit auto-commits.

    The need for this function will eventually go away once we migrate to Pandas 1.4.0.

    Args:
        df (pd.DataFrame): DataFrame to load into the SQL Table.
        name (str): name of SQL Table.
        con (sqlalchemy engine or connection): sqlalchemy.engine or sqlite3.Connection
        schema (str | None): Specify the schema (if database flavor supports this). If None, use
            default schema. Defaults to None.
        if_exists (str | None): Can be either 'fail', 'replace', or 'append'. Defaults to `fail`.
            * fail: Raise a ValueError.
            * replace: Drop the table before inserting new values.
            * append: Insert new values to the existing table.
        index (bool): Write DataFrame index as a column. Uses `index_label` as the column
            name in the table. Defaults to True.
        index_label (str | None):
            Column label for index column(s). If None is given (default) and
            `index` is True, then the index names are used.
        chunksize (int | None):
            Specify the number of rows in each batch to be written at a time.
            By default, all rows will be written at once.
        dtype (dict | int | float | bool | None):
            Specifying the datatype for columns. If a dictionary is used, the
            keys should be the column names and the values should be the
            SQLAlchemy types or strings for the sqlite3 legacy mode. If a
            scalar is provided, it will be applied to all columns.
        method (str | Callable | None):
            Controls the SQL insertion clause used:
                * None : Uses standard SQL ``INSERT`` clause (one per row).
                * 'multi': Pass multiple values in a single ``INSERT`` clause.
                * callable with signature ``(pd_table, conn, keys, data_iter)``.
    """  # noqa: E501 # FIXME CoP
    if sqlalchemy.sqlalchemy and is_version_less_than(sqlalchemy.sqlalchemy.__version__, "2.0.0"):
        with warnings.catch_warnings():
            # Note that RemovedIn20Warning is the warning class that we see from sqlalchemy
            # but using the base class here since sqlalchemy is an optional dependency and this
            # warning type only exists in sqlalchemy < 2.0.
            warnings.filterwarnings(action="ignore", category=DeprecationWarning)
            df.to_sql(
                name=name,
                con=con,
                schema=schema,
                if_exists=if_exists,  # type: ignore[arg-type] # FIXME CoP
                index=index,
                index_label=index_label,
                chunksize=chunksize,
                dtype=dtype,
                method=method,  # type: ignore[arg-type] # FIXME CoP
            )
    else:
        df.to_sql(
            name=name,
            con=con,
            schema=schema,
            if_exists=if_exists,  # type: ignore[arg-type] # FIXME CoP
            index=index,
            index_label=index_label,
            chunksize=chunksize,
            dtype=dtype,
            method=method,  # type: ignore[arg-type] # FIXME CoP
        )
